#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import datetime
from flask.cli import with_appcontext, current_app

from patientMatcher.resources import path_to_benchmark_patients
from patientMatcher.utils.add import load_demo, add_node

@click.group()
def add():
    """Add items to database using the CLI"""
    pass

@add.command()
@click.option('-id', type=click.STRING, nargs=1, required=True, help="Server ID")
@click.option('-label', type=click.STRING, nargs=1, required=True, help="Server Description")
@click.option('-token', type=click.STRING, nargs=1, required=True, help="Authorization token")
@click.option('-matching_url', type=click.STRING, nargs=1, required=True, help="URL to send match requests to")
@click.option('-accepted_content', type=click.STRING, nargs=1, required=True, help="Accepted Content-Type", default="application/vnd.ga4gh.matchmaker.v1.0+json")
@click.option('-contact', type=click.STRING, nargs=1, required=False, help="An email address")
@with_appcontext
def node(id, label, token, matching_url, accepted_content, contact=None):
    """Adds a new server to database"""
    click.echo("Adding a new MatchMaker node to database")
    node_obj = {
        '_id' : id,
        'label' : label,
        'created' : datetime.datetime.now(),
        'auth_token' : token,
        'matching_url' : matching_url,
        'accepted_content' : accepted_content,
        'contact' : contact
    }
    inserted_id, collection = add_node(mongo_db=current_app.db, obj=node_obj, is_client=False)
    if inserted_id:
        click.echo('Inserted node with ID "{}" into database collection {}'.format(inserted_id, collection))
    else:
        click.echo('Aborted')

@add.command()
@click.option('-id', type=click.STRING, nargs=1, required=True, help="Client ID")
@click.option('-token', type=click.STRING, nargs=1, required=True, help="Authorization token")
@click.option('-url', type=click.STRING, nargs=1, required=True, help="Client URL")
@click.option('-contact', type=click.STRING, nargs=1, required=False, help="Client email")
@with_appcontext
def client(id, token, url, contact=None):
    """Adds a new client to database"""
    click.echo("Adding a new client to database")
    client_obj = {
        '_id' : id,
        'created' : datetime.datetime.now(),
        'auth_token' : token,
        'base_url' : url,
        'contact' : contact
    }
    inserted_id, collection = add_node(mongo_db=current_app.db, obj=client_obj, is_client=True)
    if inserted_id:
        click.echo('Inserted client with ID "{}" into database collection {}'.format(inserted_id, collection))
    else:
        click.echo('Aborted')



@add.command()
@with_appcontext
def demodata():
    """Adds a set of 50 demo patients to database"""
    click.echo('Adding 50 test patients to database..')
    inserted_ids = load_demo(path_to_json_data=path_to_benchmark_patients, mongo_db=current_app.db,
        host=current_app.config.get('MME_HOST'))
    click.echo('inserted {} patients into db'.format(len(inserted_ids)))


add.add_command(node)
add.add_command(client)
add.add_command(demodata)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
import pymongo
from pymongo.errors import ConnectionFailure
from flask.cli import FlaskGroup, with_appcontext, current_app
from flask_mail import Message
from patientMatcher.server import create_app
from .add import add
from .remove import remove
from .update import update
cli = FlaskGroup(create_app=create_app)

@click.group()
def test():
    """Test server using CLI"""
    pass

@cli.command()
@with_appcontext
def name():
    """Returns the app name, for testing purposes, mostly"""
    app_name = current_app.name.split('.')[0]
    click.echo(app_name)
    return app_name

@cli.command()
@with_appcontext
@click.option('-recipient', type=click.STRING, nargs=1, required=True, help="Email address to send the test email to")
def email(recipient):
    """Sends a test email using config settings"""
    click.echo(recipient)

    subj = 'Test email from patientMatcher'
    body = """
        ***This is an automated message, please do not reply to this email.***<br><br>
        If you receive this email it means that email settings are working fine and the
        server will be able to send match notifications.<br>
        A mail notification will be sent when:<br>
        <ul>
            <li>A patient is added to the database and the add request triggers a search
            on external nodes producing at least one result (/patient/add endpoint).</li>

            <li>An external search is actively performed on connected nodes and returns
            at least one result (/match/external/<patient_id> endpoint).</li>

            <li>The server is interrogated by an external node and returns at least one
            result match (/match endpoint). In this case a match notification is sent to
             each contact of the result matches.</li>

            <li>An internal search is submitted to the server using a patient from the
            database (/match endpoint) and this search returns at least one match.
            In this case contact users of all patients involved will be notified
            (contact from query patient and contacts from the result patients).</li>
        </ul>
        <br>
        You can stop server notification any time by commenting the MAIL_SERVER parameter in
        config file and rebooting the server.
        <br><br>
        Kind regards,<br>
        The PatientMatcher team
    """
    kwargs = dict(subject=subj, html=body, sender=current_app.config.get('MAIL_USERNAME'), recipients=[recipient])
    message = Message(**kwargs)
    try:
        current_app.mail.send(message)
        click.echo('Mail correctly sent. Check your inbox!')
    except Exception as err:
        click.echo('An error occurred while sending test email: "{}"'.format(err))

test.add_command(name)
test.add_command(email)
cli.add_command(test)
cli.add_command(add)
cli.add_command(update)
cli.add_command(remove)



import pkg_resources

###### Files ######
hpo_filename = 'resources/hp.obo.txt'
phenotype_annotation_filename = 'resources/phenotype_annotation.tab.txt'
benchmark_patients = 'resources/benchmark_patients.json'
json_api ='resources/api.json'

###### Paths ######
path_to_hpo_terms = pkg_resources.resource_filename('patientMatcher',
    hpo_filename)

path_to_phenotype_annotations = pkg_resources.resource_filename('patientMatcher',
    phenotype_annotation_filename)

path_to_benchmark_patients = pkg_resources.resource_filename('patientMatcher',
    benchmark_patients)

path_to_json_api = pkg_resources.resource_filename('patientMatcher', json_api)



# -*- coding: utf-8 -*-

import logging
import requests

import pymongo
from flask import request

LOG = logging.getLogger(__name__)

def authorize(database, request):
    """Validate request's token against database client collection

    Args:
        database(pymongo.database.Database)
        request(request): a request object

    Returns:
        authorized(bool): True or False
    """
    token = request.headers.get('X-Auth-Token')
    query = {'auth_token' : token}
    authorized = database['clients'].find_one(query)
    if bool(authorized):
        LOG.info('Authorized client with id "{0}" submits a {1} request'.format(authorized['_id'], request.method))
    return bool(authorized)



# -*- coding: utf-8 -*-

from patientMatcher.utils.add import load_demo
from patientMatcher.match.phenotype_matcher import match, similarity_wrapper
from patientMatcher.parse.patient import mme_patient
from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations
from patient_similarity import HPO, Diseases, HPOIC, Patient
from patient_similarity.__main__ import compare_patients

PHENOTYPE_ROOT = 'HP:0000118'

def test_patient_similarity_wrapper():
    """test the wrapper around this repo: https://github.com/buske/patient-similarity"""


    # Create the information-content functionality for the HPO
    hpo = HPO(path_to_hpo_terms, new_root=PHENOTYPE_ROOT)
    assert hpo
    diseases = Diseases(path_to_phenotype_annotations)
    assert diseases
    hpoic = HPOIC(hpo, diseases, orphanet=None, patients=False, use_disease_prevalence=False,
        use_phenotype_frequency=False, distribute_ic_to_leaves=False)
    assert hpoic

    query_p_terms = ['HP:0008058', 'HP:0007033', 'HP:0002194', 'HP:0002281'] # nervous system - related HPO terms

    # test wrapper by providing same terms for query patient and match patient:
    score = similarity_wrapper(hpoic=hpoic, hpo=hpo, max_hpo_score=1.0, hpo_terms_q=query_p_terms, hpo_terms_m=query_p_terms)
    # score should be something like 0.9999999999999999
    assert round(score,12) == 1

    match_p_terms = ['HP:0008058', 'HP:0007033', 'HP:0002194']

    # test wrapper by providing almost the same terms for query patient and match patient:
    related_pheno_score = similarity_wrapper(hpoic=hpoic, hpo=hpo, max_hpo_score=1.0, hpo_terms_q=query_p_terms, hpo_terms_m=match_p_terms)
    # similarity score should be lower, but still around 1
    assert round(related_pheno_score,2) < score
    assert related_pheno_score > 0.8

    # provide completely different HPO terms for matching patient
    match_p_terms = ['HP:0003002', 'HP:0000218'] # breast cancer and high palate phenotypes
    unrelated_pheno_score = similarity_wrapper(hpoic=hpoic, hpo=hpo, max_hpo_score=1.0, hpo_terms_q=query_p_terms, hpo_terms_m=match_p_terms)
    # then unrelated_pheno_score should be almost 0
    assert round(unrelated_pheno_score,2) == 0

    # but still a positive number
    assert unrelated_pheno_score


def test_phenotype_matching(json_patients, database, demo_data_path):
    """test the algorithm that compares the phenotype of a query patient against the database"""

    # insert demo patients into test database
    inserted_ids = load_demo(demo_data_path, database, 'patientMatcher.host.se')
    assert len(inserted_ids) == 50 # 50 test cases should be loaded

    query_patient = json_patients[0]
    assert query_patient

    # this patient has HPO terms and OMIM diagnosis
    formatted_patient =  mme_patient(query_patient)
    assert len(formatted_patient['features']) > 0
    assert len(formatted_patient['disorders']) > 0

    matches_HPO_OMIM = match(database, 0.75, formatted_patient['features'],  formatted_patient['disorders'])
    assert len(matches_HPO_OMIM.keys()) == 50
    for key,value in matches_HPO_OMIM.items():
        assert 'patient_obj' in value
        assert value['pheno_score'] > 0

    features = formatted_patient['features']
    disorders = formatted_patient['disorders']
    # remove HPO terms from the query patient, test that the algorithm works anyway
    # because matching will use OMIM disorders
    formatted_patient['features'] = []
    matches_OMIM = match(database, 0.75, formatted_patient['features'],  formatted_patient['disorders'])
    assert len(matches_OMIM.keys()) > 0 and len(matches_OMIM.keys()) < 50
    for key,value in matches_OMIM.items():
        assert 'patient_obj' in value
        assert value['pheno_score'] > 0

    # remove the OMIM diagnosis from patient object. The algorithm should work
    # but it shouldn't return any match
    formatted_patient['disorders'] = []
    matches_no_phenotypes = match(database, 0.75, formatted_patient['features'],  formatted_patient['disorders'])
    assert len(matches_no_phenotypes.keys()) == 0

    # Add again features. The algorithm works again because HPO terms will be used
    formatted_patient['features'] = features
    matches_HPO = match(database, 0.75, formatted_patient['features'],  formatted_patient['disorders'])
    assert len(matches_HPO.keys()) == 50
    for key,value in matches_HPO.items():
        assert 'patient_obj' in value
        assert value['pheno_score'] > 0

    # make sure that matches obtained when OMIM and HPO terms are present are more or equal than
    # when either of these phenotype terms is present by itself
    assert len(matches_HPO_OMIM.keys()) >= len(matches_OMIM.keys())
    assert len(matches_HPO_OMIM.keys()) >= len(matches_HPO.keys())



# -*- coding: utf-8 -*-
import os
import pytest
import mongomock
from pathlib import Path
from patientMatcher.server import create_app
from patientMatcher.resources import path_to_benchmark_patients

DATABASE = 'testdb'

class MockMail:
    _send_was_called = False
    _message = None

    def send(self, message):
        self._send_was_called = True
        self._message = message

@pytest.fixture
def mock_app(database, pymongo_client):
    app = create_app()
    app.db = database
    app.client = pymongo_client
    return app

@pytest.fixture
def mock_mail():
    return MockMail()

@pytest.fixture
def mock_sender():
    return 'mock_sender'

@pytest.fixture(scope='function')
def pymongo_client(request):
    """Get a client to the mongo database"""
    mock_client = mongomock.MongoClient()
    def teardown():
        mock_client.drop_database(DATABASE)
    request.addfinalizer(teardown)
    return mock_client

@pytest.fixture(scope='function')
def database(request, pymongo_client):
    """Get an adapter connected to mongo database"""
    mongo_client = pymongo_client
    database = mongo_client[DATABASE]
    return database


@pytest.fixture(scope='function')
def demo_data_path():
    return path_to_benchmark_patients


@pytest.fixture(scope='function')
def test_client():
    """Returns a test client object"""
    client = {
        '_id' : 'client_1',
        'auth_token' : 'superSecretToken',
        'base_url' : 'http://test_node_institution.com',
        'contact' : 'test_client@email.com'
    }
    return client


@pytest.fixture(scope='function')
def test_node():
    """Returns a test client object"""
    node = {
        '_id' : 'node_1',
        'label' : 'Test node description',
        'auth_token' : 'superSecretToken',
        'matching_url' : 'http://test_node/match/',
        'accepted_content' : 'application/vnd.ga4gh.matchmaker.v1.0+json',
        'contact' : 'test_node_user@email.com'
    }
    return node


@pytest.fixture(scope='function')
def match_objs():
    """Mock the results of an internal and an external match"""
    matches = [
        {    # External match where test_patient is the query and with results
            '_id' : 'match_1',
            'has_matches' : True,
            'data' : {
                'patient' : {
                    'id' : 'P0000079',
                    'contact' : {
                        'href' : 'mailto:test_contact@email.com'
                    }
                }
            },
            'results' : [
                {
                    'node' : {'id' : 'test_node1', 'label': 'Test Node 1'},
                    'patients' : [
                        {'patient' : { 'patient_data1' : 'test_stuff1'}},
                        {'patient' : { 'patient_data2' : 'test_stuff2'}}
                    ]
                },
                {
                    'node' : {'id' : 'test_node2', 'label': 'Test Node 2'},
                    'patients' : [
                        {'patient' : { 'patient_data3' : 'test_stuff3'}}
                    ]
                }
            ],
            'match_type' : 'external'
        },
        {    # Internal match where test_patient is the query and there are no results
            '_id' : 'match_2',
            'has_matches' : False,
            'data' : {
                'patient' : {
                    'id' : 'P0000079'
                },
                'contact' : {
                    'href' : 'mailto:test_contact@email.com'
                }
            },
            'results' : [
                {
                    'node': {'id' : 'patientMatcher', 'label' : 'patientMatcher server'},
                    'patients' : [
                        {'patient' : { 'int_pat1' : 'test_stuff'}}
                    ]
                }
            ],
            'match_type' : 'internal'
        },
        {    #  Internal match where test_patient is among results
            '_id' : 'match_3',
            'has_matches' : True,
            'data' : {
                'patient' : {
                    'id' : 'external_patient_1',
                    'contact' : {
                        'href' : 'mailto:test_contact@email.com'
                    }
                }
            },
            'results' : [
                {
                    'node' : {'id' : 'test_node1', 'label': 'Test Node 1'},
                    'patients' : [
                        {'patient' : {
                            'id' : 'P0000079',
                            'contact' : {
                                'href' : 'mailto:test_contact2@email.com'
                            }
                        }}
                    ]
                }
            ],
            'match_type' : 'internal'
        },
    ]
    return matches


@pytest.fixture(scope='function')
def json_patients():
    """ returns a list containing two matchmaker-like patient objects """
    fakey_patients = [
    {
        "contact": {
          "href": "mailto:contact_email@email.com",
          "name": "A contact at an institute"
        },
        "features": [
          {
            "id": "HP:0010943",
            "label": "Echogenic fetal bowel",
            "observed": "yes"
          }
        ],
        "disorders": [
            {
                "id" : "MIM:616007"
            }
        ],
        "genomicFeatures": [
        {
            "gene": {
              "id": "LIMS2"
            },
            "type": {
              "id": "SO:0001583",
              "label": "MISSENSE"
            },
            "variant": {
              "alternateBases": "C",
              "assembly": "GRCh37",
              "end": 128412081,
              "referenceBases": "G",
              "referenceName": "2",
              "start": 128412080
            },
            "zygosity": 1
        },
        {
            "gene": {
              "id": "LIMS2"
            },
            "type": {
              "id": "SO:0001583",
              "label": "MISSENSE"
            },
            "variant": {
              "alternateBases": "A",
              "assembly": "GRCh37",
              "end": 128412067,
              "referenceBases": "G",
              "referenceName": "2",
              "start": 128412066
            },
            "zygosity": 1
        },
        {
            "gene": {
              "id": "KARS"
            },
            "type": {
              "id": "SO:0001583",
              "label": "MISSENSE"
            },
            "variant": {
              "alternateBases": "A",
              "assembly": "GRCh37",
              "end": 75665092,
              "referenceBases": "G",
              "referenceName": "16",
              "start": 75665091
            },
            "zygosity": 1
        }],
        "id": "patient_1",
        "label": "Patient number 1"
    },
    {
        "contact": {
          "href": "mailto:contact_email@email.com",
          "name": "A contact at an institute"
        },
        "features": [
          {
            "id": "HP:0001644",
            "label": "Dilated cardiomyopathy",
            "observed": "yes"
          },
        ],
        "genomicFeatures": [
          {
            "gene": {
              "id": ""
            },
            "type": {
              "id": "SO:0001583",
              "label": "MISSENSE"
            },
            "variant": {
              "alternateBases": "C",
              "assembly": "GRCh37",
              "end": 128412081,
              "referenceBases": "G",
              "referenceName": "2",
              "start": 128412080
            },
            "zygosity": 1
          },
        ],
        "id": "patient_2",
        "label": "Patient number 2"
    }]
    return fakey_patients


@pytest.fixture(scope='function')
def async_response_obj(test_node, json_patients):
    """Returns the object written to database when sending a match request to
    an async server"""

    async_response = {
        '_id' : 'async_obj_id',
        'query_id' : 'test_query_id',
        'node' : {
            'id' : test_node['_id'],
            'label' : test_node['label']
        },
        'query_patient_id' : json_patients[0]['id'],
    }
    return async_response



# -*- coding: utf-8 -*-
import requests
import json
import pymongo
from werkzeug.datastructures import Headers
from patientMatcher.utils.add import add_node, load_demo
from patientMatcher.auth.auth import authorize
from patientMatcher.server.controllers import validate_response
from patientMatcher.parse.patient import mme_patient
from patientMatcher.match.handler import patient_matches
from patientMatcher.__version__ import __version__

def test_match_async_request(mock_app, database, async_response_obj, json_patients, test_node):
    """This function tests the situation when this server is receiving a request containing
    results from an asynchronous server"""

    # send a POST request with no data to the async endpoint
    response = mock_app.test_client().post('/async_response', headers = unauth_headers())
    # server should return a request data is not valid code == 400
    assert response.status_code == 400

    # provide data object not containing a query id key
    data = { 'key1' : 'value1' }
    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())
    # server should return a not authorized response (401)
    assert response.status_code == 401

    # provide data object with query id not previously saved in database
    data = {
        'query_id' : async_response_obj['query_id'],
        'source' : 'fakey node',
        'response' : {
            "results" : [
                {
                  "score" : {
                    "patient" : 0.8
                  },
                  "patient" : json_patients[1],
                }
              ]
        }
    }

    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())
    # server should again return a not authorized response (401)
    assert response.status_code == 401

    # save async_response_obj into database
    mock_app.db['async_responses'].insert_one(async_response_obj)
    assert mock_app.db['async_responses'].find().count() == 1

    # send a response with valid data but query patient is not in database
    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())

    # server should return ok code and an error message
    assert response.status_code == 200
    resp_data = json.loads(response.data)
    assert resp_data['message'] == 'Error: could not create a valid match object from request data'

    # convert json test patient in mongodb patient object
    test_patient = mme_patient(json_patients[0])

    # save test patient in database
    mock_app.db['patients'].insert_one(test_patient)
    assert mock_app.db['patients'].find().count() == 1

    # There should be no match object in database
    assert mock_app.db['matches'].find().count() == 0

    # send a response with valid data
    # patient object is in database
    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())

    assert response.status_code == 200
    resp_data = json.loads(response.data)
    assert resp_data['message'] == 'results received, many thanks!'

    # make sure the async response entry was removed from database
    assert mock_app.db['async_responses'].find({'query_id':async_response_obj['query_id']}).count() == 0

    # and that match result was saved to server
    assert mock_app.db['matches'].find().count() == 1

    # re-introduce async response in database for further testing
    mock_app.db['async_responses'].insert_one(async_response_obj)
    # test the enpoint by providing a request with no 'response' key
    data = {
        'query_id' : async_response_obj['query_id'],
        'source' : 'fakey node',
    }
    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())
    # server should return code 400 (data is not valid)
    assert response.status_code == 400

    # test the enpoint by providing a request with 'response' key containing
    # results not conforming to the API
    data = {
        'query_id' : async_response_obj['query_id'],
        'source' : 'fakey node',
        'response' : {
            'results' : ['malformed_result1', 'malformed_result2']
        }
    }
    response = mock_app.test_client().post('/async_response',
        data=json.dumps(data), headers = unauth_headers())
    # server should return status 422 (Patient data does not conform to API)
    assert response.status_code == 422


def test_heartbeat(mock_app, database, test_client):
    """Test sending a GET request to see if app has a heartbeat"""

    # send a get request without being authorized
    response = mock_app.test_client().get('heartbeat')
    assert response.status_code == 401

    # add an authorized client to database
    ok_token = test_client['auth_token']
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # make sure that the request using its token is valid
    response = mock_app.test_client().get('heartbeat', headers = auth_headers(ok_token))
    assert response.status_code == 200

    # Make sure that all important info is returned
    data = json.loads(response.data)
    assert data['disclaimer'] == mock_app.config.get('DISCLAIMER')
    assert data['heartbeat']['version'] == __version__
    assert isinstance(data['heartbeat']['production'], bool)
    assert isinstance(data['heartbeat']['accept'], list)
    assert len(data['heartbeat']['accept']) > 0


def test_add_patient(mock_app, database, json_patients, test_client, test_node):
    """Test sending a POST request to server to add a patient"""

    patient_data = json_patients[0]
    # try to add a patient without being authorized
    response = mock_app.test_client().post('patient/add', data=json.dumps(patient_data), headers=unauth_headers())
    assert response.status_code == 401

    # add an authorized client to database
    ok_token = test_client['auth_token']

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)
    add_node(mongo_db=mock_app.db, obj=test_node, is_client=False) # add a test node, to perform external matching

    # make sure there are no matchings in 'matches' collection
    assert database['matches'].find().count()==0

    # send a malformed json object using a valid auth token
    malformed_json = "{'_id': 'patient_id' }"
    response = mock_app.test_client().post('patient/add', data=malformed_json, headers = auth_headers(ok_token))
    # and check that you get the correct error code from server(400)
    assert response.status_code == 400

    # add a patient not conforming to MME API using a valid auth token
    response = mock_app.test_client().post('patient/add', data=json.dumps(patient_data), headers = auth_headers(ok_token))
    # and check that the server returns an error 422 (unprocessable entity)
    assert response.status_code == 422

    # check that patient collection in database is empty
    assert database['patients'].find().count() == 0

    patient_obj = {'patient' : patient_data} # this is a valid patient object
    response = mock_app.test_client().post('patient/add', data=json.dumps(patient_obj), headers = auth_headers(ok_token))
    assert response.status_code == 200

    # make sure that the POST request to add the patient triggers the matching request to an external node
    assert database['matches'].find().count()==1

    # There should be one patient in database now
    assert database['patients'].find().count() == 1
    # the patient in database has label "Patient number 2"
    assert database['patients'].find({'label' : 'Patient number 1'}).count() == 1

    # Add same patient again and see that label is unchanged and there is still one patient in database:
    patient_obj = {'patient' : patient_data} # this is a valid patient object
    response = mock_app.test_client().post('patient/add', data=json.dumps(patient_obj), headers = auth_headers(ok_token))
    assert response.status_code == 200
    assert database['patients'].find().count() == 1
    assert database['patients'].find({'label' : 'Patient number 1'}).count() == 1

    # modify patient label and check the update command (add a patient with the same id) works
    patient_data['label'] = 'modified patient label'
    patient_obj = {'patient' : patient_data}
    response = mock_app.test_client().post('patient/add', data=json.dumps(patient_obj), headers = auth_headers(ok_token))
    assert response.status_code == 200

    # There should still be one patient in database
    assert database['patients'].find().count() == 1

    # But its label is changed
    assert database['patients'].find({'label' : 'Patient number 1'}).count() == 0
    assert database['patients'].find({'label' : 'modified patient label'}).count() == 1

    # make sure that the POST request to add modify patient triggers the matching request to an external node again
    assert database['matches'].find().count()==3


def test_metrics(mock_app, database, test_client, demo_data_path, match_objs):
    """Testing viewing the list of patients on server for authorized users"""

    # send a get request without being authorized
    response = mock_app.test_client().get('metrics')
    assert response.status_code == 401

    # add an authorized client to database
    ok_token = test_client['auth_token']

    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # make sure that database contains at least one client with auth_token == test_client['auth_token']
    clients = mock_app.db['clients'].find({'auth_token' : ok_token }).count()
    assert clients > 0

    # load demo data of 50 test patients
    inserted_ids = load_demo(demo_data_path, database, mock_app.config.get('MME_HOST'))
    assert len(inserted_ids) == 50 # 50 test cases should be loaded

    # load mock matches into database
    database.matches.insert_many(match_objs)
    assert database.matches.find().count() == 3

    # if a valid token is provided the server should return metrics with patient data and matching results
    auth_response = mock_app.test_client().get('metrics', headers = auth_headers(ok_token))
    assert auth_response.status_code == 200

    data = json.loads(auth_response.data)
    assert data['disclaimer'] # disclaimer should be returned
    metrics = data['metrics']

    assert metrics['numberOfCases'] == 50
    assert metrics['numberOfSubmitters'] > 0
    assert metrics['numberOfGenes'] > metrics['numberOfUniqueGenes']
    assert metrics['numberOfVariants'] > metrics['numberOfUniqueVariants']
    assert metrics['numberOfFeatures'] > metrics['numberOfUniqueFeatures']
    assert metrics['numberOfCasesWithDiagnosis'] >0
    assert metrics['numberOfUniqueGenesMatched'] == 0 # no gene was provided in match_obj results
    assert metrics['numberOfRequestsReceived'] == 2 # Sent 2 requests
    assert metrics['numberOfPotentialMatchesSent'] == 1 # Just one has returned results


def test_nodes_view(mock_app, database, test_node, test_client):
    """testing viewing the list of connected nodes as an authenticated client"""

    # insert a test client in database
    add_node(mongo_db=database, obj=test_client, is_client=True)
    ok_token = test_client['auth_token']

    # If you try to see nodes without being authorized
    response = mock_app.test_client().get('nodes')
    # You should get a not authorized code from server
    assert response.status_code == 401

    # since there are no connected nodes in database
    assert database['nodes'].find().count() == 0
    # When you send an authorized request
    response = mock_app.test_client().get('nodes', headers = auth_headers(ok_token))
    data = json.loads(response.data)
    # you shoud get an empty list
    assert data == []

    # insert a test node in database
    add_node(mongo_db=database, obj=test_node, is_client=False)
    # When you send an authorized request
    response = mock_app.test_client().get('nodes', headers = auth_headers(ok_token))
    data = json.loads(response.data)
    # this time you should get a list with one element
    assert len(data) == 1
    # and the id of the element is the id of the node
    assert data[0]['id'] == test_node['_id']


def test_delete_patient(mock_app, database, demo_data_path, test_client, match_objs):
    """Test deleting a patient from database by sending a DELETE request"""

    # load demo data to mock database using function located under utils/load
    inserted_ids = load_demo(demo_data_path, database, mock_app.config.get('MME_HOST'))
    assert len(inserted_ids) == 50 # 50 test cases should be loaded

    # 50 cases present on patients collection
    assert database['patients'].find().count() == 50
    delete_id = 'P0000079'

    # try to delete patient without auth token:
    response = mock_app.test_client().delete(''.join(['patient/delete/', delete_id]))
    assert response.status_code == 401

    # Add a valid client node
    ok_token = test_client['auth_token']
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # Send delete request providing a valid token but a non valid id
    response = mock_app.test_client().delete(''.join(['patient/delete/', 'not_a_valid_ID']), headers = auth_headers(ok_token))
    assert response.status_code == 200
    data = json.loads(response.data)
    # but server returns error
    assert data['message'] == 'ERROR. Could not delete a patient with ID not_a_valid_ID from database'

    assert database['matches'].find().count() == 0 # no matches in database
    # insert into database some mock matching objects
    database['matches'].insert_many(match_objs)

    # patient "delete_id" should have two associated matches in database
    assert database['matches'].find({'data.patient.id' : delete_id}).count() == 2

    # Send valid patient ID and valid token
    response = mock_app.test_client().delete(''.join(['patient/delete/', delete_id]), headers = auth_headers(ok_token))
    assert response.status_code == 200

    # make sure that the patient was removed from database
    assert database['patients'].find().count() == 49

    # make sure that patient matches are also gone
    assert database['matches'].find().count() == 1


def test_patient_matches(mock_app, database, match_objs, test_client):
    """testing the endpoint that retrieves the matchings by patient ID"""

    # Add a valid client node
    ok_token = test_client['auth_token']
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    # start from a database with no matches
    assert database['matches'].find().count() == 0
    # import mock matches into datababase
    database['matches'].insert_many(match_objs)
    # database now should have two matching objects

    # test endpoint to get matches by ID
    # test by sending a non-authorized request
    response = mock_app.test_client().get('matches/P0000079')
    # response gives a 401 code (not authorized)
    assert response.status_code == 401

    # try with an authorized request with a used ID that is not in database
    response = mock_app.test_client().get('matches/unknown_patient', headers = auth_headers(ok_token))
    # response gives success
    assert response.status_code == 200
    data = json.loads(response.data)
    # but the patient is not found by server
    assert data['message'] == 'Could not find any matches in database for patient ID unknown_patient'

    # Try with authenticates request and valid patient
    response = mock_app.test_client().get('matches/P0000079', headers = auth_headers(ok_token))
    # response gives success
    assert response.status_code == 200
    data = json.loads(response.data)
    # and there are matches in it
    assert len(data['matches']) == 2 # 2 matches returned because endpoint returns only matches with results

    # Test that there are actually 3 matches by calling directly the function returning matches
    matches = patient_matches(database=database, patient_id='P0000079', type=None, with_results=False)
    assert len(matches) == 3

    # Call the same function to get only external matches
    matches = patient_matches(database=database, patient_id='P0000079', type='external', with_results=False)
    assert len(matches) == 1

    # Call the same function to get only external matches
    matches = patient_matches(database=database, patient_id='P0000079', type='internal', with_results=False)
    assert len(matches) == 2


def test_match(mock_app, json_patients, test_client, demo_data_path, database):
    """Testing patient matching against patientMatcher database (internal match)"""

    # add an authorized client to database
    ok_token = test_client['auth_token']
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True)

    query_patient = {'patient' : json_patients[0]}

    # load demo data in mock database
    inserted_ids = load_demo(demo_data_path, database, mock_app.config.get('MME_HOST'))

    # test the API response validator with non valid patient data:
    malformed_match_results = {'results': 'fakey_results'}
    assert validate_response(malformed_match_results) == 422

    # make sure that there are no patient matches in the 'matches collection'
    assert database['matches'].find().count()==0

    # send a POST request to match patient with patients in database
    response = mock_app.test_client().post('/match', data=json.dumps(query_patient), headers = auth_headers(ok_token))
    assert response.status_code == 200 # POST request should be successful
    data = json.loads(response.data)
    # data should contain results and the max number of results is as defined in the config file
    assert len(data['results']) == mock_app.config['MAX_RESULTS']

    assert type(data['results']) == list # which is a list
    assert 'patient' in data['results'][0] # of patients
    assert 'score' in data['results'][0] # with matching scores

    # make sure that there are match object is created in db for this internal matching
    assert database['matches'].find().count()==1


def test_match_external(mock_app, test_client, test_node, database, json_patients):
    """Testing the view that is sending post request to trigger matches on external nodes"""

    # add an authorized client to database
    ok_token = test_client['auth_token']
    add_node(mongo_db=mock_app.db, obj=test_client, is_client=True) # required to trigger external matches

    a_patient = json_patients[0]
    parsed_patient = mme_patient(a_patient)

    # insert patient into mock database:
    assert database['patients'].find().count() == 0
    inserted_id = database['patients'].insert_one(parsed_patient).inserted_id
    assert database['patients'].find().count() == 1

    # send an un-authorized match request to server
    response = mock_app.test_client().post(''.join(['/match/external/', inserted_id]))
    # server should return 401 (not authorized)
    assert response.status_code == 401

    # send an authorized request with a patient ID that doesn't exist on server:
    response = mock_app.test_client().post(''.join(['/match/external/', 'not_a_valid_ID']), headers = auth_headers(ok_token))
    # Response is valid
    assert response.status_code == 200
    data = json.loads(response.data)
    # but server returns error
    assert data['message'] == 'ERROR. Could not find any patient with ID not_a_valid_ID in database'

    # there are no matches in mock database
    assert database['matches'].find().count() == 0
    # after sending an authorized request with a patient ID that exists on database

    # Check that external matching doesn't work if there are no connected nodes:
    response = mock_app.test_client().post(''.join(['/match/external/', inserted_id]), headers = auth_headers(ok_token))
    assert response.status_code == 200
    data = json.loads(response.data)
    assert data['message'] == 'Could not find any other node connected to this MatchMaker server'

    # Try to send a request for a match on a node that does not exist
    response = mock_app.test_client().post(''.join(['/match/external/', inserted_id, '?node=meh']), headers = auth_headers(ok_token))
    assert response.status_code == 200
    data = json.loads(response.data)
    # And check that node not found is in response message
    assert data['message'] == 'ERROR. Could not find any connected node with id meh in database'

    # insert a connected node
    add_node(mongo_db=mock_app.db, obj=test_node, is_client=False) # required for external matches
    # send a request to match patients against all nodes
    response = mock_app.test_client().post(''.join(['/match/external/', inserted_id]), headers = auth_headers(ok_token))

    # Response should be valid
    assert response.status_code == 200
    # And a new match should be created in matches collection
    assert database['matches'].find().count() == 1

    # send a request to match patients against the specific existing node:
    response = mock_app.test_client().post(''.join(['/match/external/', inserted_id, '?node=', test_node['_id']]), headers = auth_headers(ok_token))
    # Response should be valid
    assert response.status_code == 200

    # And a new match should be created in matches collection. So total matches are 2
    assert database['matches'].find().count() == 2


def unauth_headers():
    head = {
        'Content-Type': 'application/vnd.ga4gh.matchmaker.v1.0+json',
        'Accept': ['application/vnd.ga4gh.matchmaker.v1.0+json', 'application/json'],
        'X-Auth-Token': 'wrong_token'
    }
    return head

def auth_headers(test_token):
    head = {
        'Content-Type': 'application/vnd.ga4gh.matchmaker.v1.0+json',
        'Accept': ['application/vnd.ga4gh.matchmaker.v1.0+json', 'application/json'],
        'X-Auth-Token': test_token
    }
    return head



# -*- coding: utf-8 -*-

from patientMatcher.utils.add import load_demo
from patientMatcher.utils.patient import patients
import pymongo

def test_patients(demo_data_path, database):
    """Tests the function that retrieves patient objects"""

    # load demo data in mock database
    inserted_ids = load_demo(demo_data_path, database, False)
    assert type(inserted_ids) == list
    assert len(inserted_ids) == 50 # 50 test cases are loaded

    # get all 50 patients available in database
    all_patients = list(patients(database=database))
    assert len(all_patients) == 50

    # test collecting just a few patients:
    selected_patients = list(patients(database=database, ids=inserted_ids[:3]))
    assert len(inserted_ids[:3]) == 3
    assert len(selected_patients) == 3



# -*- coding: utf-8 -*-

import sys
import enlighten
import json
import logging
from pymongo import MongoClient

from patientMatcher.parse.patient import mme_patient
from patientMatcher.match.handler import external_matcher

LOG = logging.getLogger(__name__)

def load_demo(path_to_json_data, mongo_db, host):
    """Inserts demo patient data into database
        Demo data consists of a set of 50 patients from this paper: http://onlinelibrary.wiley.com/doi/10.1002/humu.22850

        Args:
            path_to_demo_data(str): absolute path to json file containing the demo patients.
            mongo_db(pymongo.database.Database)
            host(str): MME_HOST parameter in config file

        Returns:
            inserted_ids(list): the database ID of the inserted patients
    """
    patients = [] # a list of dictionaries
    inserted_ids = []

    #open json file and try to insert one patient at the time
    try:
        LOG.info('reading patients file')
        with open(path_to_json_data) as json_data:
            patients = json.load(json_data)
            # create a progress bar
            pbar = enlighten.Counter(total=len(patients), desc='', unit='patients')
            for json_patient in patients:

                #parse patient into format accepted by database
                patient = mme_patient(json_patient)

                inserted_id = backend_add_patient(mongo_db=mongo_db, host=host, patient=patient)[1]
                if inserted_id:
                    inserted_ids.append(inserted_id)
                pbar.update()

    except Exception as err:
        LOG.fatal("An error occurred while importing benchmarking patients: {}".format(err))

    return inserted_ids


def backend_add_patient(mongo_db, host, patient, match_external=False):
    """
        Insert or update a patient in patientMatcher database

        Args:
            mongo_db(pymongo.database.Database)
            host(str): MME_HOST parameter in config file
            patient(dict) : a MME patient entity

        Returns:
            inserted_id(str) : the ID of the inserted patient or None if patient couldn't be saved
    """
    modified = None
    upserted = None
    matching_obj = None

    try:
        result = mongo_db['patients'].replace_one({'_id': patient['_id']}, patient , upsert=True)
        modified = result.modified_count
        upserted = result.upserted_id

    except Exception as err:
        LOG.fatal("Error while inserting a patient into database: {}".format(err))

    # this will happen only if patient is added via POST request,
    # and if there is a change in patients' collections (new patient or updated patient)
    # Matching is not triggered by inserting demo data into database
    if match_external and (modified or upserted):
        matching_obj = external_matcher(mongo_db, host, patient)
        #save matching object to database
        mongo_db['matches'].insert_one(matching_obj)

    return modified, upserted, matching_obj


def add_node(mongo_db, obj, is_client):
    """
        Insert a new node (client or server) into the database

        Args:
            mongo_db(pymongo.database.Database)
            obj(dict): a client or a server object to add to database
            is_client(bool): if True the new node a client of this server, if False this server is a client of the new the node

        Returns:
            inserted_id(str), collection(str): a tuple with values inserted_id and collection name
    """
    inserted_id = None
    collection = None

    if is_client:
        collection = "clients"
    else:
        collection = "nodes"

    try:
        inserted_id = mongo_db[collection].insert_one(obj).inserted_id
    except Exception as err:
        LOG.fatal('Error while inserting a new client/server node to database:{}'.format(err))

    return inserted_id, collection



# -*- coding: utf-8 -*-
import logging
from flask_mail import Message

LOG = logging.getLogger(__name__)


def notify_match_internal(database, match_obj, admin_email, mail, notify_complete):
    """Send an email to patient contacts after an internal match

    Args:
        database(pymongo.database.Database): patientMatcher database
        match_obj(dict): an object containing both query patient(dict) and matching results(list)
        admin_email(str): email of the server admin
        mail(flask_mail.Mail): an email instance
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    # Internal matching can be triggered by a patient in the same database or by a patient on a connected node.
    # In the first case notify both querier contact and contacts in the result patients.
    # in the second case notify only contacts from patients in the results list.
    sender = admin_email
    patient_id = None
    patient_label = None
    results = None
    recipient = None
    email_subject = 'MatchMaker Exchange: new patient match available.'
    email_body = None

    # check if query patient already belongs to patientMatcher database:
    internal_patient = database['patients'].find_one({'_id':match_obj['data']['patient']['id']})
    if internal_patient:
        #If patient used for the search is in patientMatcher database, notify querier as well:
        patient_id = match_obj['data']['patient']['id']
        patient_label = match_obj['data']['patient'].get('label')
        recipient = match_obj['data']['patient']['contact']['href'][7:]
        email_body = active_match_email_body(patient_id=patient_id, match_results=match_obj['results'], patient_label=patient_label,
            external_match=False, notify_complete=notify_complete)
        LOG.info('Sending an internal match notification for query patient with ID:{0}. Patient contact: {1}'.format(patient_id, recipient))

        kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
        message = Message(**kwargs)
        # send email using flask_mail
        try:
            mail.send(message)
        except Exception as err:
            LOG.error('An error occurred while sending an internal match notification: {}'.format(err))

    # Loop over the result patients and notify their contact about the matching with query patient
    for result in match_obj['results'][0]['patients']: #this list has only one element since there is only one internal node

        patient_id = result['patient']['id']

        # do not notify when patient in results is the same as the one used for query
        if internal_patient and internal_patient['_id'] == patient_id:
            continue

        patient_label =  result['patient'].get('label')
        recipient = result['patient']['contact']['href'][7:]
        email_body = passive_match_email_body(patient_id, match_obj['data']['patient'], patient_label, notify_complete)
        LOG.info('Sending an internal match notification for match result with ID {}'.format(patient_id))

        kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
        message = Message(**kwargs)
        # send email using flask_mail
        try:
            mail.send(message)
        except Exception as err:
            LOG.error('An error occurred while sending an internal match notification: {}'.format(err))


def notify_match_external(match_obj, admin_email, mail, notify_complete):
    """Send an email to patients contacts to notify a match on external nodes

    Args:
        match_obj(dict): an object containing both query patient(dict) and matching results(list)
        admin_email(str): email of the server admin
        mail(flask_mail.Mail): an email instance
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    sender = admin_email
    patient_id = match_obj['data']['patient']['id']
    patient_label = match_obj['data']['patient'].get('label')
    recipient = match_obj['data']['patient']['contact']['href'][7:]
    email_subject = 'MatchMaker Exchange: new patient match available.'
    email_body = active_match_email_body(patient_id=patient_id, match_results=match_obj['results'], patient_label=patient_label,
        external_match=True, notify_complete=notify_complete)
    LOG.info('Sending an external match notification for query patient with ID {0}. Patient contact: {1}'.format(patient_id, recipient))

    kwargs = dict(subject=email_subject, html=email_body, sender=sender, recipients=[recipient])
    message = Message(**kwargs)
    # send email using flask_mail
    try:
        mail.send(message)
    except Exception as err:
        LOG.error('An error occurred while sending an external match notification: {}'.format(err))


def active_match_email_body(patient_id, match_results, patient_label=None, external_match=False, notify_complete=False):
    """Returns the body message of the notification email when the patient was used as query patient

    Args:
        patient_id(str): the ID of the patient submitted by the  MME user which will be notified
        match_results(list): a list of patients which match with the patient whose contact is going to be notified
        external_match(bool): True == match in connected nodes, False == match with other patients in database
        patient_label(str): the label of the patient submitted by the  MME user which will be notified (not mandatory field)
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email

    Returns:
        html(str): the body message
    """
    search_type = 'against the internal database of MatchMaker patients'
    if external_match:
        search_type = 'against external nodes connected to MatchMaker'

    html = """
        ***This is an automated message, please do not reply to this email.***<br><br>
        <strong>MatchMaker Exchange patient matching notification:</strong><br><br>
        Patient with ID <strong>{0}</strong>, label <strong>{1}</strong>.
        This search returned these potential matches</strong>:<br>
        <strong>{2}</strong><br>
        You might directly contact the matching part using the address specified in patient's data or review matching
        results in the portal you used to submit your patient.
        <br><br>
        Kind regards,<br>
        The PatientMatcher team
    """.format(patient_id, patient_label, html_format(match_results, 0, notify_complete))

    return html


def passive_match_email_body(patient_id, matched_patient, patient_label=None, notify_complete=False):
    """Returns the body message of the notification email when the patient was used as query patient

    Args:
        patient_id(str): the ID of the patient submitted by the MME user which will be notified
        matched_patient(dict): a patient object
        patient_label(str): the label of the patient submitted by the  MME user which will be notified (not mandatory field)
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email

    Returns:
        html(str): the body message
    """
    html = """
        ***This is an automated message, please do not reply.***<br>
        <strong>MatchMaker Exchange patient matching notification:</strong><br><br>
        Patient with <strong>ID {0}</strong>,<strong> label {1}</strong> was recently returned as a match result
        in a search performed using a patient with these specifications:<br>
        <strong>{2}</strong><br>
        You might directly contact the matching part using the address specified in patient's data or review matching
        results in the portal you used to submit your patient.
        <br><br>
        Kind regards,<br>
        The PatientMatcher team
    """.format(patient_id, patient_label, html_format(matched_patient, 0, notify_complete))

    return html


def html_format(obj, indent=0, notify_complete=False):
    """Formats one or more patient objects to a nice html string

    Args:
        obj(list): a list of patient objects or a patient object
        notify_complete(bool): set to False to NOT notify variants and phenotype terms by email
    """
    if isinstance(obj, list): # a list pf match results
        htmls = []
        for k in obj:
            htmls.append(html_format(obj=k, indent=indent+1, notify_complete=notify_complete))

        return '[<div style="margin-left: %dem">%s</div>]' % (indent, ',<br>'.join(htmls))

    if isinstance(obj, dict): # patient object
        htmls = []
        for k,v in obj.items():
            if notify_complete or k in ['node', 'patients', 'patient', 'contact', 'id', 'name', 'href', 'institution']:
                htmls.append("<span style='font-style: italic; color: #888'>%s</span>: %s" % (k,html_format(obj=v,indent=indent+1,
                    notify_complete=notify_complete)))

        return '{<div style="margin-left: %dem">%s</div>}' % (indent, ',<br>'.join(htmls))

    return str(obj)



# -*- coding: utf-8 -*-
import logging
from datetime import date

LOG = logging.getLogger(__name__)

def general_metrics(db):
    """Create an object with database metrics

    Args:
        db(pymongo.database.Database)

    Returns:
        metrics(dict): According to the MME API it should be a dictionary like this:
            {
                "metrics": {
                    "numberOfCases": 0,
                    "numberOfSubmitters": 0,
                    "numberOfGenes": 0,
                    "numberOfUniqueGenes": 0,
                    "numberOfVariants": 0,
                    "numberOfUniqueVariants": 0,
                    "numberOfFeatures": 0,
                    "numberOfUniqueFeatures": 0,
                    "numberOfFeatureSets": 0, # endpoint is not returning this, at the moment
                    "numberOfUniqueGenesMatched": 0,
                    "numberOfCasesWithDiagnosis": 0,
                    "numberOfRequestsReceived": 0,
                    "numberOfPotentialMatchesSent": 0,
                    "dateGenerated": "2017-08-24",

                },
                "disclaimer": "Disclaimer text...",
                "terms": "Terms text..."
            }
    """
    # get gene/occurrence for all genes in db
    n_genes = 0
    gene_occurrs = item_occurrence(db, 'genomicFeatures', 'genomicFeatures.gene', 'genomicFeatures.gene.id')
    for gene_count in gene_occurrs:
        n_genes += gene_count['count']

    # get numberOfUniqueVariants/occurrence for all variants in db
    variant_occurr = item_occurrence(db, 'genomicFeatures', 'genomicFeatures.variant', 'genomicFeatures.variant')
    n_vars = 0
    for var in variant_occurr:
        n_vars += var.get('count')

    # get feature/occurrence for all features in db
    n_feat = 0
    feat_occurr = item_occurrence(db, 'features', 'features.id')
    for feat in feat_occurr:
        n_feat += feat.get('count')

    # include in unique_gene_matches only matches actively returned by the server (internal)
    match_type = {'match_type':'internal'}
    unique_gene_matches = db.matches.distinct('results.patients.patient.genomicFeatures.gene', match_type)

    metrics = {
        'numberOfCases' : db.patients.find().count(),
        'numberOfSubmitters' : len(db.patients.distinct('contact.href')),
        'numberOfGenes' : n_genes,
        'numberOfUniqueGenes': len(db.patients.distinct('genomicFeatures.gene')),
        'numberOfVariants' : n_vars,
        'numberOfUniqueVariants' : len(db.patients.distinct('genomicFeatures.variant')),
        'numberOfFeatures' : n_feat,
        'numberOfUniqueFeatures' : len(db.patients.distinct('features.id')),
        'numberOfUniqueGenesMatched' : len(unique_gene_matches),
        'numberOfCasesWithDiagnosis' : db.patients.find({'disorders': {'$exists': True, '$ne' : []} }).count(),
        'numberOfRequestsReceived' : db.matches.find({'match_type':'internal'}).count(),
        'numberOfPotentialMatchesSent' : db.matches.find({'match_type':'internal', 'has_matches': True}).count(),
        'dateGenerated' : str(date.today())
    }
    return metrics


def item_occurrence(db, unw1, group, unw2=None):
    """Get a list of item/occurrence in patient collection

    Args:
        db(pymongo.database.Database)
        unw1(string): first nested unwind item
        group(string): item to group results by
        unw2(string): second nested unwind item # none if nested level is missing

    Returns:
        item_occurr(list) example: [{'id':'item_obj', 'count': item_occurrence}, ..]
    """
    # create query pipeline
    pipeline = [{"$unwind": ''.join(['$',unw1])}]
    if unw2:
        pipeline.append({"$unwind": ''.join(['$',unw2])})
    pipeline.append({"$group": {"_id": ''.join(['$',group]), "count": {"$sum": 1}}})
    item_occurr = list(db.patients.aggregate(pipeline))
    return item_occurr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import logging
from patientMatcher.parse.patient import features_to_hpo, disorders_to_omim
from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations
from patient_similarity import HPO, Diseases, HPOIC, Patient
from patient_similarity.__main__ import compare_patients

LOG = logging.getLogger(__name__)
PHENOTYPE_ROOT = 'HP:0000001'

def match(database, max_score, features, disorders):
    """Handles phenotype matching algorithm

    Args:
        database(pymongo.database.Database)
        max_score(float): a number between 0 and 1
        features(list): a list of phenotype feature objects (example ID = HP:0008619)
        disorders(list): a list of OMIM diagnoses (example ID = MIM:616007 )

    Returns:
        matches(dict): a dictionary of patient matches with phenotype matching score
    """
    matches = {}

    hpo_terms = []
    omim_terms = []
    query_fields = []

    hpoic = None
    hpo = None

    LOG.info('\n\n###### Running phenotype matcher module ######')

    if features: # at least one HPO term is specified
        hpo_terms = features_to_hpo(features)
        # compare against all cases which also have features (HPO terms)
        query_fields.append({'features': {'$exists': True, '$ne': []}})

        # Create the information-content functionality for the HPO
        LOG.info('Creating HPO information content')
        hpo = HPO(path_to_hpo_terms, new_root=PHENOTYPE_ROOT)
        diseases = Diseases(path_to_phenotype_annotations)
        hpoic = HPOIC(hpo, diseases, orphanet=None, patients=False,
                      use_disease_prevalence=False,
                      use_phenotype_frequency=False,
                      distribute_ic_to_leaves=False)

    if disorders: # at least one OMIM term was provided
        omim_terms = disorders_to_omim(disorders)
        query_fields.append({'disorders.id': {"$in" : omim_terms}})

    # build a database query taking into account patient features (HPO terms) and disorders (omim)
    if len(query_fields) > 0:
        query = { '$or' : query_fields }
        LOG.info('Searching for patients in database with the following query:{}'.format(query))
        pheno_matching_patients = list(database['patients'].find(query))
        LOG.info("\n\nFOUND {} patients matching patients's phenotype tracts\n\n".format(len(pheno_matching_patients)))

        for i in range(len(pheno_matching_patients)):
            patient = pheno_matching_patients[i]
            LOG.info('## Evaluating phenotype similarity with patient {} ##'.format(i+1))
            similarity = evaluate_pheno_similariy(hpoic, hpo, hpo_terms, omim_terms, patient,
                max_score)

            match = {
                'patient_obj' : patient,
                'pheno_score' : similarity,
            }
            matches[patient['_id']] = match

    return matches


def evaluate_pheno_similariy(hpoic, hpo, hpo_terms, disorders, pheno_matching_patient, max_similarity):
    """Evaluates the similarity of two patients based on phenotype features

        Args:
            hpoic(class) : the information content for the HPO
            hpo(class): an instance of the class for interacting with HPO
            hpo_terms(list): HPO terms of the query patient
            disorders(list): OMIM disorders of the query patient
            pheno_matching_patient(patient_obj): a patient object from the database
            max_similarity(float): a floating point number representing the highest value allowed for a feature

        Returns:
            patient_similarity(float): the computed phenotype similarity among the patients
    """
    patient_similarity = 0
    hpo_score = 0
    omim_score = 0

    max_omim_score = 0
    max_hpo_score = 0

    # get matching patients HPO terms as a list
    matching_hpo_terms = features_to_hpo(pheno_matching_patient.get('features'))
    matching_omim_terms = disorders_to_omim(pheno_matching_patient.get('disorders'))

    # If both query patient and matching patient contain features to compare (HPO terms)
    if hpo_terms and matching_hpo_terms:
        LOG.info('HPO terms available for comparison')
        # If both query patient and matching patient contain OMIM diagnoses
        if disorders and matching_omim_terms:
            LOG.info('OMIM diagnoses available for comparison')
            max_omim_score = max_similarity/2
            max_hpo_score = max_similarity/2

        else: # OMIM diagnoses are missing --> HPO score represents max similarity
            max_hpo_score = max_similarity

        hpo_score = similarity_wrapper(hpoic, hpo, max_hpo_score, hpo_terms, matching_hpo_terms)

    else: # HPO terms missing
        # similarity is computed using OMIM terms,
        # Penalty for missing HPO terms: max_omim_score = max_similarity/2
        LOG.debug('Missing HPO terms, phenotype comparison based on OMIM diagnoses.')
        max_omim_score = max_similarity/2

    if max_omim_score: # OMIM terms can be compared
        omim_score = evaluate_subcategories(disorders, matching_omim_terms, max_omim_score)

    patient_similarity = hpo_score + omim_score
    LOG.info('patient phenotype score: {0} (OMIM:{1}, HPO:{2})'.format(patient_similarity,
        omim_score, hpo_score))
    return patient_similarity


def similarity_wrapper(hpoic, hpo, max_hpo_score, hpo_terms_q, hpo_terms_m):
    """A wrapper around patient-similarity repository:
    https://github.com/buske/patient-similarity.

    Args:
        hpoic(class) : the information content for the HPO
        hpo(class): an instance of the class for interacting with HPO
        max_hpo_score(float): max score which can be assigned to HPO similarity
        hpo_terms_q(list): a list of HPO terms from query patient
        hpo_terms_m(list): a list of HPO terms from match patient

    Returns:
        score(float): simgic similarity score after HPO term comparison
    """
    # create Patient object from query patient data:
    terms = set()
    for term_id in hpo_terms_q:
        term = hpo[term_id]
        if term:
            terms.add(term)
    query_patient = Patient(id='q', hp_terms=terms)

    # create Patient object from match patient data:
    terms = set()
    for term_id in hpo_terms_m:
        term = hpo[term_id]
        if term:
            terms.add(term)
    match_patient = Patient(id='m', hp_terms=terms)

    # Get simgic similarity score for HPO terms comparison
    # Range is 0 to 1, with 0=no similarity and 1=highest similarity
    score_obj = compare_patients(hpoic=hpoic, patient1=query_patient,
        patient2=match_patient, scores=['simgic'])
    simgic_score = score_obj.get('simgic')
    LOG.info('patient-similarity module returned a simgic score of {}'.format(simgic_score))
    relative_simgic_score = simgic_score * max_hpo_score
    return relative_simgic_score


def evaluate_subcategories(list1, list2, max_score):
    """returns a numerical representation of the similarity of two lists of strings

        Args:
            list1(list): a list of strings (this is a list of items from the query patient)
            list2(list): another list of strings (list of items from the patients in database)
            max_score(float): the maximum value to return if the lists are identical

        Returns:
            matching_score(float): a number reflecting the similarity between the lists
    """
    matching_score = 0
    if len(list1)>0:
        list_item_score = max_score/len(list1) # the max value of each matching item between lists
        n_shared_items = len(set(list1).intersection(list2)) # number of elements shared between the lists
        matching_score = n_shared_items * list_item_score
    return matching_score



# -*- coding: utf-8 -*-

import logging
from bson import json_util
import json
from flask import Blueprint, request, current_app, jsonify
from flask_negotiate import consumes, produces

from patientMatcher.utils.add import backend_add_patient
from patientMatcher.utils.notify import notify_match_external, notify_match_internal
from patientMatcher.auth.auth import authorize
from patientMatcher.match.handler import internal_matcher, patient_matches, async_match
from patientMatcher.parse.patient import validate_api, mme_patient
from patientMatcher.constants import STATUS_CODES
from . import controllers

LOG = logging.getLogger(__name__)
blueprint = Blueprint('server', __name__)
API_MIME_TYPE = 'application/vnd.ga4gh.matchmaker.v1.0+json'

@blueprint.route('/patient/add', methods=['POST'])
@consumes(API_MIME_TYPE, 'application/json')
@produces('application/json')
def add():
    """Add patient to database"""

    formatted_patient = controllers.check_request(current_app.db, request)
    if isinstance(formatted_patient, int): # some error must have occurred during validation
        return controllers.bad_request(formatted_patient)

    match_external = False
    if controllers.get_nodes(database=current_app.db):
        match_external = True

    # else import patient to database
    modified, inserted, matching_obj= backend_add_patient(mongo_db=current_app.db, patient=formatted_patient,
        match_external=match_external, host=current_app.config.get('MME_HOST'))
    message = {}

    if modified:
        message['message'] = 'Patient was successfully updated.'
    elif inserted:
        message['message'] = 'Patient was successfully inserted into database.'
    else:
        message['message'] = 'Database content is unchanged.'

    # if patient is matching any other patient on other nodes
    # and match notifications are on
    if current_app.config.get('MAIL_SERVER') and matching_obj and len(matching_obj.get('results')):
        # send an email to patient's contact:
        notify_match_external(match_obj=matching_obj, admin_email=current_app.config.get('MAIL_USERNAME'),
            mail=current_app.mail, notify_complete=current_app.config.get('NOTIFY_COMPLETE'))

    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route('/patient/delete/<patient_id>', methods=['DELETE'])
def delete(patient_id):
    #check if request is authorized
    resp = None
    if not authorize(current_app.db, request): # not authorized, return a 401 status code
        message = STATUS_CODES[401]['message']
        resp = jsonify(message)
        resp.status_code = 401
        return resp

    LOG.info('Authorized client is removing patient with id {}'.format(patient_id))
    message = controllers.delete_patient(current_app.db, patient_id)
    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route('/heartbeat', methods=['GET'])
def heartbeat():
    """Get the server specs"""
    resp = None
    if authorize(current_app.db, request):
        LOG.info('Authorized client requests heartbeat..')
        LOG.info('current_app is {}'.format(current_app))
        disclaimer = current_app.config.get('DISCLAIMER')
        result = controllers.heartbeat(disclaimer)
        resp = jsonify(result)
        resp.status_code = 200

    else: # not authorized, return a 401 status code
        return controllers.bad_request(401)

    return resp


@blueprint.route('/metrics', methods=['GET'])
def metrics():
    """Get database metrics"""
    resp = None
    if authorize(current_app.db, request):
        LOG.info('Authorized client requests metrics..')
        results = controllers.metrics(database=current_app.db)
        resp = jsonify({'metrics' : results, 'disclaimer' : current_app.config.get('DISCLAIMER')})
        resp.status_code = 200

    else: # not authorized, return a 401 status code
        return controllers.bad_request(401)

    return resp


@blueprint.route('/nodes', methods=['GET'])
def nodes():
    """Get a list of all nodes connected to this MME server"""
    resp = None
    if not authorize(current_app.db, request):
        message = STATUS_CODES[401]['message']
        resp = jsonify(message)
        resp.status_code = 401
        return resp

    LOG.info('Authorized client requests list of connected nodes..')
    results = controllers.get_nodes(database=current_app.db)
    resp = jsonify(results)
    resp.status_code = 200
    return resp


@blueprint.route('/matches/<patient_id>', methods=['GET'])
def matches(patient_id):
    """Get all matches (external and internal) for a patient ID"""
    LOG.info('getting all matches for patient {}'.format(patient_id))
    resp = None
    message = {}
    if not authorize(current_app.db, request): # not authorized, return a 401 status code
        message = STATUS_CODES[401]['message']
        resp = jsonify(message)
        resp.status_code = 401
        return resp

    # return only matches with at least one result
    results = patient_matches(current_app.db, patient_id)
    if results:
        message = json.loads(json_util.dumps({'matches' : results}))
    else:
        message['message'] = "Could not find any matches in database for patient ID {}".format(patient_id)
    resp = jsonify(message)
    resp.status_code = 200
    return resp


@blueprint.route('/match/external/<patient_id>', methods=['POST'])
def match_external(patient_id):
    """Trigger a patient matching on external nodes by providing a patient ID"""
    resp = None
    message = {}
    if not authorize(current_app.db, request): # not authorized, return a 401 status code
        message['message'] = STATUS_CODES[401]['message']
        resp = jsonify(message)
        resp.status_code = 401
        return resp

    LOG.info('Authorized clients is matching patient with ID {} against external nodes'.format(patient_id))
    query_patient = controllers.patient(current_app.db, patient_id)

    if not query_patient:
        LOG.error('ERROR. Could not find any patient with ID {} in database'.format(patient_id))
        message['message'] = "ERROR. Could not find any patient with ID {} in database".format(patient_id)
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    node = request.args.get('node')

    # if search should be performed on a specific node, make sure node is in database
    if node and not current_app.db['nodes'].find({'_id':node}).count():
        LOG.info('ERROR, theres no node with id "{}" in database'.format(request.args['node']))
        message['message'] = 'ERROR. Could not find any connected node with id {} in database'.format(request.args['node'])
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    host = current_app.config.get('MME_HOST') # Introduce yourself in request
    matching_obj = controllers.match_external(current_app.db, host, query_patient, node)

    if not matching_obj:
        message['message'] = "Could not find any other node connected to this MatchMaker server"
        resp = jsonify(message)
        resp.status_code = 200
        return resp

    results = matching_obj.get('results')

    # if patient is matching any other patient on other nodes
    # and match notifications are on
    if current_app.config.get('MAIL_SERVER') and matching_obj and len(results):
        # send an email to patient's contact:
        notify_match_external(match_obj=matching_obj, admin_email=current_app.config.get('MAIL_USERNAME'),
            mail=current_app.mail, notify_complete=current_app.config.get('NOTIFY_COMPLETE'))

    resp = jsonify({'results':results})
    resp.status_code = 200
    return resp


@blueprint.route('/async_response', methods=['POST'])
def asynch_match_response():
    """Receives and handles asynchronous match responses.
    This implies that patientMatcher has already submitted a match request
    to a server x and has received not results, but a query id in response.
    Query id was previously saved into patientMatcher database.
    This delayed response from server x contains match results and
    the same query id and server x source key.
    """
    LOG.info('Receiving an asynchronous request.')
    data = controllers.check_async_request(current_app.db, request)
    # returned data after check_async_request is a dictionary if request is valid
    if isinstance(data, int): # some error must have occurred during validation
        return controllers.bad_request(data)

    # create a match object from specifics you first sent to the node
    async_match_obj = async_match(current_app.db, data)
    if async_match_obj:
        # save match object to database
        LOG.info('saving match object to database')
        current_app.db['matches'].insert_one(async_match_obj)

        # Remove query_id from async responses, this is not used any more
        current_app.db['async_responses'].delete_one({'query_id':data['query_id']})
    else:
        message = "Error: could not create a valid match object from request data"
        resp = jsonify({'message' : message})
        resp.status_code = 200
        return resp

    # notify matches if there are matching patients and app is configured to do so
    if len(async_match_obj['results']) > 0 and current_app.config.get('MAIL_SERVER'):
        # send an email to patient's contact:
        notify_match_external(match_obj=async_match_obj, admin_email=current_app.config.get('MAIL_USERNAME'),
            mail=current_app.mail, notify_complete=current_app.config.get('NOTIFY_COMPLETE'))

    resp = jsonify({'message':'results received, many thanks!'})
    resp.status_code = 200
    return resp


@blueprint.route('/match', methods=['POST'])
@consumes(API_MIME_TYPE)
@produces(API_MIME_TYPE, 'application/json')
def match_internal():
    """Match a query patient against patients from local database and
    returns a response containing patients which are most similar to the
    query patient"""

    query_patient = controllers.check_request(current_app.db, request)
    if isinstance(query_patient, int): # some error must have occurred during validation
        return controllers.bad_request(query_patient)

    LOG.info('Matching a query patient against database')

    max_pheno_score = current_app.config.get('MAX_PHENO_SCORE', 0.25) # get max pheno score from app settings, if available
    max_geno_score = current_app.config.get('MAX_GT_SCORE', 0.75) # get max genotyping score from app settings, if available
    max_results = current_app.config.get('MAX_RESULTS')

    # get a list of matching patients ordered by score
    match_obj = internal_matcher(current_app.db, query_patient, max_pheno_score, max_geno_score, max_results)
    # save matching object to database
    current_app.db['matches'].insert_one(match_obj)
    matches = match_obj['results'][0]['patients'] #results[0] because there is just one node (internal match)

    # if notifications are on and there are matching results
    if current_app.config.get('MAIL_SERVER') and len(matches):
        notify_match_internal(database=current_app.db, match_obj=match_obj,
            admin_email=current_app.config.get('MAIL_USERNAME'), mail=current_app.mail,
            notify_complete=current_app.config.get('NOTIFY_COMPLETE'))


    validate_response = controllers.validate_response({'results': matches})
    if isinstance(validate_response, int): # some error must have occurred during validation
        return controllers.bad_request(validate_response)

    LOG.info('Returning {} matching patients from database'.format(len(matches)))

    # return response with results
    validate_response["disclaimer"] = current_app.config.get('DISCLAIMER')
    resp = jsonify(validate_response)
    resp.status_code = 200
    return resp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import datetime
import requests
import pymongo
import json
from werkzeug.datastructures import Headers
from patientMatcher.match.genotype_matcher import match as genomatch
from patientMatcher.match.phenotype_matcher import match as phenomatch
from patientMatcher.parse.patient import json_patient

LOG = logging.getLogger(__name__)


def patient_matches(database, patient_id, type=None, with_results=True):
    """Retrieve all matches for a patient specified by an ID

    Args:
        database(pymongo.database.Database)
        patient_id(str): ID of a patient in database
        type(str): type of matching, external or internal
        with_results(bool): if True only matches with results are returned

    Returns:
        matches(list): a list of dictionaries = [ {match_obj1}, {match_obj2}, .. ]
    """
    query = {
        '$or' : [
            {'data.patient.id' : patient_id }, # collect matches triggered by patient
            {'results.patients.patient.id' : patient_id} # and matches where patient is among results
        ]}
    if type:
        query['match_type'] = type
    if with_results:
        query['has_matches'] = True

    matches = list(database['matches'].find(query))
    LOG.info(matches)
    return matches


def internal_matcher(database, patient_obj, max_pheno_score, max_geno_score, max_results=5):
    """Handles a query patient matching against the database of patients

    Args:
        database(pymongo.database.Database)
        patient_obj(dic): a mme formatted patient object
        max_pheno_score(float): a number between 0 and 1
        max_geno_score(float): a number between 0 and 1
        max_results(int): the maximum number of results that the server should return.

    Returns:
        internal_match(dict): a matching object with results(list) sorted by score
    """
    json_pat = json_patient(patient_obj)
    pheno_matches = []
    pheno_m_keys = []
    geno_matches = []
    geno_m_keys = []
    matches = []

    # phenotype score can be obtained if patient has an associated phenotype (HPO or OMIM terms)
    if patient_obj.get('features') or patient_obj.get('disorders'):
        LOG.info('Matching phenotypes against database patients..')
        pheno_matches = phenomatch(database, max_pheno_score, patient_obj.get('features',[]), patient_obj.get('disorders',[]))
        pheno_m_keys = list(pheno_matches.keys())

    # genomic score can be obtained if patient has at least one genomic feature
    if patient_obj.get('genomicFeatures') and len(patient_obj['genomicFeatures']) > 0:
        LOG.info('Matching variants/genes against database patients..')
        geno_matches = genomatch(database, patient_obj['genomicFeatures'], max_geno_score)
        geno_m_keys = list(geno_matches.keys())

    # obtain unique list of all patient IDs returned by the 2 algorithms:
    unique_patients = list(set(pheno_m_keys + geno_m_keys))

    # create matching result objects with combined score from the 2 algorithms
    for key in unique_patients:
        pheno_score = 0
        geno_score = 0
        patient_obj = None

        if key in pheno_m_keys:
            pheno_score = pheno_matches[key]['pheno_score']
            patient_obj = pheno_matches[key]['patient_obj']

        if key in geno_m_keys:
            geno_score = geno_matches[key]['geno_score']
            patient_obj = geno_matches[key]['patient_obj']

        p_score = pheno_score + geno_score
        score = {
            'patient' : p_score,
            '_genotype' : geno_score,
            '_phenotype' : pheno_score
        }
        match = {
            'patient' : json_patient(patient_obj),
            'score' : score
        }
        if score['patient']>0:
            matches.append(match)

    # sort patient matches by patient (combined) score
    sorted_matches = sorted(matches, key=lambda k : k['score']['patient'], reverse=True)

    # this is saved to server, regardless of the results returned by the nodes
    has_matches = False
    if sorted_matches:
        has_matches = True

    internal_match = {
        'created' : datetime.datetime.now(),
        'has_matches' : has_matches,
        'data' : {'patient' : json_pat}, # description of the patient submitted
        'results' : [{
            'node' : { 'id': 'patientMatcher', 'label': 'patientMatcher server'},
            'patients' : sorted_matches[:max_results]
        }],
        'match_type' : 'internal'
    }

    # return matching object
    return internal_match


def external_matcher(database, host, patient, node=None):
    """Handles a query patient matching against all connected MME nodes

    Args:
        database(pymongo.database.Database)
        host(str): Name of this server (MME_HOST in config file)
        patient(dict) : a MME patient entity
        node(str): id of the node to search in

    Returns:
        external_match(dict): a matching object containing a list of results in 'results' field
    """
    query = {}
    if node:
        query['_id'] = node
    #get all connected nodes or ther one specified by user
    connected_nodes = list(database['nodes'].find(query))
    if len(connected_nodes) == 0:
        LOG.error("Could't find any connected MME nodes. Aborting external matching.")
        return None

    # create request headers
    headers = Headers()
    data = {'patient': json_patient(patient)} # convert into something that follows the API specs

    # this is saved to server, regardless of the results returned by the nodes
    external_match = {
        'created' : datetime.datetime.now(),
        'has_matches' : False, # it changes if a similar patient is returned by any other MME nodes
        'data' : data, # description of the patient submitted
        'results' : [],
        'errors' : [],
        'match_type' : 'external',
    }

    LOG.info("Matching patient against {} node(s)..".format(len(connected_nodes)))
    for node in connected_nodes:

        server_name = node['_id']
        node_url = node['matching_url']
        token = node['auth_token']
        request_content_type = node['accepted_content']

        headers = {'Content-Type': request_content_type,
            'Accept': 'application/vnd.ga4gh.matchmaker.v1.0+json',
            "X-Auth-Token": token,
            'Host' : host
        }
        LOG.info('sending HTTP request to server: "{}"'.format(server_name))
        # send request and get response from server
        json_response = None
        server_return = None
        try:
            server_return = requests.request(
                method = 'POST',
                url = node_url,
                headers = headers,
                data = json.dumps(data)
            )
            json_response = server_return.json()
        except Exception as json_exp:
            error = json_exp
            LOG.error('Server returned error:{}'.format(error))

            error_obj = {
                'node' : { 'id': node['_id'], 'label' : node['label'] },
                'error' : str(error)
            }
            external_match['errors'].append(error_obj)

        if json_response:
            LOG.info('server returns the following response: {}'.format(json_response))

            result_obj = {
                'node' : { 'id': node['_id'], 'label' : node['label'] },
                'patients' : []
            }

            # node returned results
            if 'results' in json_response:
                results = json_response['results']
                if len(results):
                    external_match['has_matches'] = True
                    for result in results:
                        result_obj['patients'].append(result)

                    external_match['results'].append(result_obj)

            elif 'query_id' in json_response: # node is sending a query id (asynchronous response)
                #  query id must be saved in database because the node will be sending
                # a delayed request with results and this query_id as identifier
                LOG.info('Node {} is sending an async response, saving query id to server'.format(server_name))
                save_async_response(database=database, node_obj=node, query_id=json_response['query_id'],
                    query_patient_id=patient['_id'])
            else:
                LOG.error('JSON response from server was:{}'.format(json_response))

    return external_match


def async_match(database, response_data):
    """Creates a match object from data received from an asyc server

    Args:
        database(pymongo.database.Database)
        response_data(dict): data provided by async server in follow up request

    Returns:
        async_match(dict): a match object
    """
    # get query id you received from async server when you sent your request
    query_id = response_data['query_id'] # can't be None, checked upstream
    # collect data saved in database when you first received the async response
    async_response = database['async_responses'].find_one({'query_id':query_id})
    patient_id = async_response.get('query_patient_id')
    patient_obj = database['patients'].find_one({'_id':patient_id})
    node = async_response.get('node')
    if not patient_obj:
        LOG.error("Couldn't find a patient with id '{} in database.".format(patient_id))
        return
    if not node:
        LOG.error("Couldn't find node '{}' in database ".format(node.get('label')))
        return
    # response key is not None, checked upstream
    resp = response_data.get('response')
    results_obj = {
        'node' : node,
        'patients' : resp.get('results')
    }

    async_match = {
        'created' : datetime.datetime.now(),
        'has_matches' : True if len(results_obj.get('patients'))>0 else False,
        'data' : {'patient' : patient_obj},
        'errors' : [], # This can be modified if I know that async server returns error
        'match_type' : 'external',
        'async' : True,
        'results' : [results_obj]
    }
    return async_match


def save_async_response(database, node_obj, query_id, query_patient_id):
    """Creates an async response object in the async_responses collection.
    This object stores info about the query patient sent to an async server and the
    query_id received from it. Async server does provide match results in a
    subsequent request, along with the same query_id string, to trace it back to the
    original request initiated by patienrMatcher.

    Args:
        database(pymongo.database.Database)
        node_obj(dict): an async node object
        query_id(str): an id returned by async server in its response
        query_patient_id(str): id of internal patient matched against patients
            in async server
    """
    async_response = {
        'query_id' : query_id,
        'query_patient_id' : query_patient_id,
        'node' : {
            'id' : node_obj['_id'],
            'label' : node_obj.get('label')
        },
        'created' : datetime.datetime.now()
    }
    database['async_responses'].insert_one(async_response)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import requests
from clint.textui import progress
import click
from patientMatcher.constants import PHENOTYPE_TERMS


@click.group()
def update():
    """Update patientMatcher resources"""
    pass

@update.command()
@click.option('--test',
              help='Use this flag to test the function',
              is_flag=True)
def resources(test):
    """Updates HPO terms and disease ontology from the web.
    Specifically collect files from:
    http://purl.obolibrary.org/obo/hp.obo
    http://compbio.charite.de/jenkins/job/hpo.annotations/lastStableBuild/artifact/misc/phenotype_annotation.tab
    """
    files = {}
    for key,item in PHENOTYPE_TERMS.items():
        url = item['url']
        destination = item['resource_path']
        r = requests.get(url, stream=True)
        total_length = int(r.headers.get('content-length'))
        if test: # read file and get its size
            files[key] = total_length # create an object for each downloadable file and save its length
            if total_length:
                click.echo('file {} found at the requested URL'.format(key))
            continue
        with open(destination, 'wb') as f: #overwrite file
            for chunk in progress.bar(r.iter_content(chunk_size=1024), expected_size=(total_length/1024) + 1):
                if chunk:
                    f.write(chunk)
                    f.flush()



# -*- coding: utf-8 -*-
import logging
from flask import jsonify
from jsonschema import ValidationError
from patientMatcher.constants import STATUS_CODES
from patientMatcher.utils.stats import general_metrics
from patientMatcher.utils.delete import delete_by_query
from patientMatcher.utils.patient import patients
from patientMatcher.parse.patient import json_patient, validate_api, mme_patient
from patientMatcher.auth.auth import authorize
from patientMatcher.match.handler import external_matcher
from patientMatcher.__version__ import __version__

LOG = logging.getLogger(__name__)

def heartbeat(disclaimer):
    """Return a heartbeat as defined here:https://github.com/ga4gh/mme-apis/blob/master/heartbeat-api.md"""

    hbeat = {
        "heartbeat": {
            "production": True,
            "version": __version__,
            "accept": ["application/vnd.ga4gh.matchmaker.v1.0+json", "application/vnd.ga4gh.matchmaker.v1.1+json"]
        },
        "disclaimer": disclaimer,
    }
    return hbeat


def metrics(database):
    """return database metrics"""
    db_metrics = general_metrics(database)
    return db_metrics


def get_nodes(database):
    """Get all connected nodes as a list of objects with node_id and node_label as elements"""
    results = list(database['nodes'].find())
    nodes = []
    for node in results:
        nodes.append( { 'id': node['_id'], 'description': node['label']} )
    return nodes


def patient(database, patient_id):
    """Return a mme-like patient from database by providing its ID"""
    query_patient = None
    query_result = list(patients(database, ids=[patient_id]))
    if query_result:
        query_patient = query_result[0]
    return query_patient


def match_external(database, host, query_patient, node=None):
    """Trigger an external patient matching for a given patient object"""
    # trigger the matching and save the matching id to variable
    matching_obj = external_matcher(database, host, query_patient, node)
    # save matching object to database
    if matching_obj:
        database['matches'].insert_one(matching_obj)
    return matching_obj


def check_request(database, request):
    """Check if request is valid, if it is return MME formatted patient
       Otherwise return error code.
    """
    check_result = None

    # check that request is using a valid auth token
    if not authorize(database, request):
        LOG.info("Request is not authorized")
        return 401

    try: # make sure request has valid json data
        request_json = request.get_json(force=True)
    except Exception as err:
        LOG.info("Json data in request is not valid:{}".format(err))
        return 400

    try: # validate json data against MME API
        validate_api(json_obj=request_json, is_request=True)
    except Exception as err:
        LOG.info("Patient data does not conform to API:{}".format(err))
        return 422

    formatted_patient = mme_patient(json_patient=request_json['patient'])
    return formatted_patient


def check_async_request(database, request):
    """Check if an asynchronous request is valid.
    Basically json data must be valid and the query ID should be
    already present in async responses database collection"""

    data = None
    try: # Check if request contains valid data
        data = request.json
        LOG.info('Request data looks valid. Source is {}'.format(data.get('source')))
    except:
        LOG.error('Request data is not valid. Abort')
        return 400

    # check if query ID was previously saved into async responses collection
    query_id = data.get('query_id')
    if query_id:
        async_response = database['async_responses'].find_one({'query_id':query_id})
        LOG.info('Async response is {}'.format(async_response))
    if query_id is None or async_response is None:
        LOG.error('Async request not authorized. Abort')
        return 401

    resp = data.get('response')
    if resp is None:
        LOG.error("Async server did not provide any 'response' object")
        return 400
    try: # validate json response (results)
        validate_api(json_obj=resp, is_request=False)
    except Exception as err:
        LOG.info("Patient data does not conform to API:{}".format(err))
        return 422

    return data


def validate_response(matches):
    """Validates patient matching results before sending them away in a response"""

    try: # validate json data against MME API
        validate_api(json_obj=matches, is_request=False)
    except ValidationError as err:
        LOG.info("Patient data does not conform to API:{}".format(err))
        return 422
    return matches


def bad_request(error_code):
    """Crete an automatic response based on custom error codes"""
    message = STATUS_CODES[error_code]['message']
    resp = jsonify(message)
    resp.status_code = error_code
    return resp


def delete_patient(database, patient_id):
    """Remove a patient by ID"""
    message = ''

    # first delete all matches in database for this patient:
    query = {'data.patient.id' : patient_id}
    deleted = delete_by_query(query, database, 'matches')
    LOG.info('deleted {} matche/s triggered by this patient'.format(deleted))


    query = {'_id' : patient_id}
    deleted = delete_by_query(query, database, 'patients')
    message = {}
    if deleted == 1:
        message['message'] = 'Patient and its matches were successfully deleted from database'
    else:
        message['message'] = 'ERROR. Could not delete a patient with ID {} from database'.format(patient_id)
    return message



# -*- coding: utf-8 -*-

import json
from jsonschema import validate, RefResolver, FormatChecker
from pkgutil import get_data
import logging

LOG = logging.getLogger(__name__)
SCHEMA_FILE = 'api.json'

def mme_patient(json_patient):
    """
        Accepts a json patient and converts it to a MME patient,
        formatted as required by patientMatcher database

        Args:
            patient_obj(dict): a patient object as in https://github.com/ga4gh/mme-apis

        Returns:
            mme_patient(dict) : a mme patient entity
    """

    mme_patient = {
        '_id' : json_patient['id'],
        'label' : json_patient.get('label'),
        'sex' : json_patient.get('sex'),
        'contact' : json_patient['contact'],
        'features' : json_patient.get('features'),
        'genomicFeatures' : json_patient.get('genomicFeatures'),
        'disorders' : json_patient.get('disorders'),
        'species' : json_patient.get('species'),
        'ageOfOnset' : json_patient.get('ageOfOnset'),
        'inheritanceMode' : json_patient.get('inheritanceMode')
    }

    # remove keys with empty values from mme_patient object
    mme_patient = {k:v for k,v in mme_patient.items() if v is not None}

    return mme_patient


def json_patient(mme_patient):
    """ Converts a mme patient into a json-like as in the MME APIs

        Args:
            mme_patient(dict): a patient object as it is stored in database

        Returns:
            json_patient(dict): a patient object conforming to MME API
    """
    json_patient = mme_patient
    if '_id' in json_patient:
        json_patient['id'] = json_patient['_id']
        json_patient.pop('_id')

    return json_patient


def features_to_hpo(features):
    """Extracts HPO terms from a list of phenotype features of a patient

        Args:
            features(list): a list of features dictionaries

        Returns:
            hpo_terms(list): a list of HPO terms. Example : ['HP:0100026', 'HP:0009882', 'HP:0001285']
    """

    hpo_terms = [feature.get('_id') for feature in features if feature.get('_id')]
    if len(hpo_terms) == 0:
        hpo_terms = [feature.get('id') for feature in features if feature.get('id')]
    return hpo_terms


def disorders_to_omim(disorders):
    """Extracts OMIM terms from a list of disorders of a patient

        Args:
            disorders(list): a list of disorders

        Returns:
            omim_terms(list): a list of OMIM terms. Example : ['MIM:616007', 'MIM:614665']
    """

    omim_terms = [disorder.get('id') for disorder in disorders if disorder.get('id')]
    return omim_terms


def gtfeatures_to_genes(gtfeatures):
    """Extracts all gene names from a list of genomic features

    Args:
        gtfeatures(list): a list of genomic features objects

    Returns:
        gene_set(list): a list of unique gene names contained in the features
    """
    genes = []
    for feature in gtfeatures:
        if 'gene' in feature and feature['gene']['id']: # collect non-null gene IDs
            genes.append(feature['gene']['id'])
    gene_set = list(set(genes))
    return gene_set


def gtfeatures_to_variants(gtfeatures):
    """Extracts all variants from a list of genomic features

    Args:
        gtfeatures(list): a list of genomic features objects

    Returns:
        variants(list): a list of variants
    """
    variants = []
    for feature in gtfeatures:
        if 'variant' in feature:
            variants.append(feature['variant'])

    return variants


def validate_api(json_obj, is_request):
    """Validates a request against the MME API
       The code for the validation against the API and the API specification is taken from
       this project: https://github.com/MatchmakerExchange/reference-server

    Args:
        json_obj(dict): json request or response to validate
        is_request(bool): True if it is a request, False if it is a response

    Returns
        validated(bool): True or False
    """
    validated = True
    schema = '#/definitions/response'
    if is_request:
        schema = '#/definitions/request'

    LOG.info("Validating against SCHEMA {}".format(schema))

    # get API definitions
    schema_data = json.loads(get_data('patientMatcher.resources', SCHEMA_FILE).decode('utf-8'))
    resolver = RefResolver.from_schema(schema_data)
    format_checker = FormatChecker()
    resolver_schema = resolver.resolve_from_url(schema)
    validate(json_obj, resolver_schema, resolver=resolver, format_checker=format_checker)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import click
from flask.cli import with_appcontext, current_app

from patientMatcher.utils.delete import delete_by_query

@click.group()
def remove():
    """Remove items from database using the CLI"""
    pass

@remove.command()
@click.option('-id', type=click.STRING, nargs=1, required=False, help="ID of the patient to be removed from database")
@click.option('-label', type=click.STRING, nargs=1, required=False, help="Label of the patient to be removed from database")
@click.option('-remove_matches/-leave_matches', default=False, help="Remove or leave on db matches triggered by patient")
@with_appcontext
def patient(id, label, remove_matches):
    """Removing a patient from patientMatcher providing its ID"""

    if not id and not label:
        click.echo('Error: either ID and/or label should be provided to delete a patient.')
        raise click.Abort()

    if remove_matches and not id:
        click.echo('Please provide patient ID and not label to remove all its matches.')
        raise click.Abort()

    query = {}
    if id:
        query['_id'] = id
    if label:
        query['label'] = label

    n_removed = delete_by_query(query=query, mongo_db= current_app.db, mongo_collection='patients')
    click.echo('Number of patients removed from database:{}'.format(n_removed))

    if remove_matches:
        # this will remove ONLY matches where this patient was the query patient
        # NOT those where patient was among the matching results
        query = {'data.patient.id' : id}
        n_removed = delete_by_query(query=query, mongo_db= current_app.db, mongo_collection='matches')
        click.echo('Number of matches for this patient removed from database:{}'.format(n_removed))


@remove.command()
@click.option('-id', type=click.STRING, nargs=1, required=True, help="ID of the client to be removed from database")
@with_appcontext
def client(id):
    """Remove a client from database by providing its ID"""

    query = {'_id' : id}
    n_removed = delete_by_query(query=query, mongo_db= current_app.db, mongo_collection='clients')
    click.echo('Number of clients removed from database:{}'.format(n_removed))


@remove.command()
@click.option('-id', type=click.STRING, nargs=1, required=True, help="ID of the node to be removed from database")
@with_appcontext
def node(id):
    """Remove a node from database by providing its ID"""

    query = {'_id' : id}
    n_removed = delete_by_query(query=query, mongo_db= current_app.db, mongo_collection='nodes')
    click.echo('Number of nodes removed from database:{}'.format(n_removed))



# -*- coding: utf-8 -*-

from patientMatcher.utils.add import load_demo, backend_add_patient
from patientMatcher.parse.patient import mme_patient
from patientMatcher.match.handler import internal_matcher, save_async_response, external_matcher

def test_internal_matching(demo_data_path, database, json_patients):
    """Testing the combined matching algorithm"""

    # load demo data to mock database using function located under utils/load
    inserted_ids = load_demo(demo_data_path, database, False)
    assert len(inserted_ids) == 50 # 50 test cases are loaded

    # format test patient for query:
    test_mme_patients = [ mme_patient(patient) for patient in json_patients ]

    a_patient = test_mme_patients[0]
    assert a_patient

    match_obj = internal_matcher(database, a_patient, 0.5, 0.5)
    matches = match_obj['results'][0]['patients']
    assert len(matches) > 0

    higest_scored_patient = matches[0]
    lowest_scored_patient = matches[-1]

    assert higest_scored_patient['score']['patient'] > lowest_scored_patient['score']['patient']


def test_external_matching(database, test_node, json_patients):
    """Testing the function that trigger patient matching across connected nodes"""

    patient = json_patients[0]

    # insert test node object in database
    database['nodes'].insert_one(test_node)

    # insert patient object in database
    inserted_ids = backend_add_patient(mongo_db=database, host='patientMatcher.host.se', patient=patient, match_external=False)
    assert inserted_ids

    ext_m_result = external_matcher(database, 'pmatcher', patient, test_node['_id'])
    assert isinstance(ext_m_result, dict)
    assert ext_m_result['data']['patient']['id'] == patient['id']
    assert ext_m_result['has_matches'] == False
    assert ext_m_result['match_type'] == 'external'


def test_save_async_response(database, test_node):
    """Testing the function that saves an async response object to database"""

    # Test database should not contain async responses
    assert database['async_responses'].find().count() == 0

    # Save an async response using the matching handler
    save_async_response(database=database, node_obj=test_node,
        query_id='test', query_patient_id='test_patient')

    # async_responses database collection should now contain one object
    async_response = database['async_responses'].find_one()
    assert async_response['query_id'] == 'test'
    assert async_response['query_patient_id'] == 'test_patient'
    assert async_response['node']['id'] == test_node['_id']
    assert async_response['node']['label'] == test_node['label']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pymongo
from flask_mail import Message

from patientMatcher.cli.commands import cli
from patientMatcher.utils.add import load_demo

def test_appname(mock_app):
    runner = mock_app.test_cli_runner()
    result = runner.invoke(cli, ['test', 'name'])
    assert result.output == 'patientMatcher\n'


def test_sendemail(mock_app, mock_mail):
    mock_app.mail = mock_mail
    mock_app.config['MAIL_USERNAME'] = 'sender@email.com'

    runner = mock_app.test_cli_runner()
    # When invoking the test email command with a recipient paramrter
    result = runner.invoke(cli, ['test', 'email', '-recipient', 'test_user@mail.com'])

    # Make sure that mock mail send method was called and mock email is sent
    assert mock_mail._send_was_called
    assert mock_mail._message
    assert 'Mail correctly sent' in result.output


def test_cli_add_node(mock_app, database, test_node):
    # make sure that "nodes" collection is empty
    assert database['nodes'].find().count() == 0

    # test add a server using the app cli
    runner = mock_app.test_cli_runner()
    result =  runner.invoke(cli, ['add', 'node', '-id', test_node['_id'],
        '-label', 'This is a test node', '-token', test_node['auth_token'],
        '-matching_url', test_node['matching_url'],'-accepted_content',
        test_node['accepted_content']])
    assert result.exit_code == 0
    assert 'Inserted node' in result.output

    # check that the server was added to the "nodes" collection
    assert database['nodes'].find().count() == 1

    # Try adding the node again
    result =  runner.invoke(cli, ['add', 'node', '-id', test_node['_id'],
        '-label', 'This is a test node', '-token', test_node['auth_token'],
        '-matching_url', test_node['matching_url'],'-accepted_content',
        test_node['accepted_content']])
    assert result.exit_code == 0
    # And you should get an abort message
    assert 'Aborted' in result.output
    # And number of nodes in database should stay the same
    assert database['nodes'].find().count() == 1

def test_cli_add_client(mock_app, database, test_client):

    # make sure that "clients" collection is empty
    assert database['client'].find().count() == 0

    # test add a server using the app cli
    runner = mock_app.test_cli_runner()
    result =  runner.invoke(cli, ['add', 'client', '-id', test_client['_id'],
        '-token', test_client['auth_token'], '-url', test_client['base_url']])

    assert result.exit_code == 0
    assert 'Inserted client' in result.output

    # check that the server was added to the "nodes" collection
    assert database['clients'].find().count() == 1

    # Try adding the client again
    result =  runner.invoke(cli, ['add', 'client', '-id', test_client['_id'],
        '-token', test_client['auth_token'], '-url', test_client['base_url']])
    assert result.exit_code == 0
    # And you should get an abort message
    assert 'Aborted' in result.output
    # And number of clients in database should stay the same
    assert database['clients'].find().count() == 1


def test_cli_remove_client(mock_app, database, test_client):

    # Add a client to database
    runner = mock_app.test_cli_runner()
    result =  runner.invoke(cli, ['add', 'client', '-id', test_client['_id'],
        '-token', test_client['auth_token'], '-url', test_client['base_url']])
    assert result.exit_code == 0

    # check that the server was added to the "nodes" collection
    assert database['clients'].find().count() == 1

    # Use the cli to remove client
    result =  runner.invoke(cli, ['remove', 'client', '-id', test_client['_id'] ])

    # check that command is executed withour errors
    assert result.exit_code == 0

    # and that client is gone from database
    assert database['clients'].find().count() == 0


def test_cli_remove_node(mock_app, database, test_node):

    # Add a node to database
    runner = mock_app.test_cli_runner()
    result =  runner.invoke(cli, ['add', 'node', '-id', test_node['_id'],
        '-label', 'This is a test node', '-token', test_node['auth_token'],
        '-matching_url', test_node['matching_url'],'-accepted_content',
        test_node['accepted_content']])
    assert result.exit_code == 0

    # check that the server was added to the "nodes" collection
    assert database['nodes'].find().count() == 1

    # Use the cli to remove client
    result =  runner.invoke(cli, ['remove', 'node', '-id', test_node['_id'] ])

    # check that command is executed withour errors
    assert result.exit_code == 0

    # and that node is gone from database
    assert database['nodes'].find().count() == 0


def test_cli_update_resources(mock_app):

    runner = mock_app.test_cli_runner()

    # run resources update command with --test flag:
    result =  runner.invoke(cli, ['update', 'resources'])
    assert result.exit_code == 0


def test_cli_add_demo_data(mock_app, database):

    runner = mock_app.test_cli_runner()

    # make sure that "patients" collection is empty
    assert database['patients'].find().count() == 0

    # run the load demo command without the -compute_phenotypes flag
    result =  runner.invoke(cli, ['add', 'demodata'])
    assert result.exit_code == 0

    # check that the 50 demo patients where inserted into database
    assert database['patients'].find().count() == 50


def test_cli_remove_patient(mock_app, database, json_patients, match_objs):

    runner = mock_app.test_cli_runner()

    # add a test patient to database
    test_patient = json_patients[0]
    test_patient['_id'] = 'P0000079'
    inserted_id = mock_app.db['patients'].insert_one(test_patient).inserted_id
    assert inserted_id == 'P0000079'

    # there is now 1 patient in database
    assert database['patients'].find().count() == 1

    # test that without a valid id or label no patient is removed
    result =  runner.invoke(cli, ['remove', 'patient', '-id', '', '-label', ''])
    assert 'Error' in result.output

    # Add mock patient matches objects to database
    database['matches'].insert_many(match_objs)
    # There should be 2 matches in database for this patient:
    assert database['matches'].find( {'data.patient.id' : inserted_id }).count() == 2

    # involke cli command to remove the patient by id and label
    result =  runner.invoke(cli, ['remove', 'patient', '-id', inserted_id, '-label', 'Patient number 1', '-leave_matches'])
    assert result.exit_code == 0

    # check that the patient was removed from database
    assert database['patients'].find().count() == 0

    # But matches are still there
    assert database['matches'].find( {'data.patient.id' : inserted_id }).count() == 2

    # Run remove patient command with option to remove matches but without patient ID
    result =  runner.invoke(cli, ['remove', 'patient', '-label', 'Patient number 1', '-remove_matches'])
    # And make sure that it doesn't work
    assert 'Please provide patient ID and not label to remove all its matches.' in result.output

    # Test now the proper command to remove patient matches:
    result =  runner.invoke(cli, ['remove', 'patient', '-id', inserted_id, '-remove_matches'])
    assert result.exit_code == 0

    # And make sure that patient removal removed its matchings
    assert database['matches'].find( {'data.patient.id' : inserted_id }).count() == 0



#Turns on debugging features in Flask
DEBUG = True

#secret key:
SECRET_KEY = 'MySuperSecretKey'

# Database connection string
DB_USERNAME = 'pmUser'
DB_PASSWORD = 'pmPassword'
DB_NAME = 'pmatcher'
DB_HOST = '127.0.0.1'
DB_PORT = 27017
DB_URI = "mongodb://{}:{}@{}:{}/{}".format(DB_USERNAME, DB_PASSWORD, DB_HOST, DB_PORT, DB_NAME)

# Matching Algorithms scores
# sum of MAX_GT_SCORE and MAX_PHENO_SCORE should be 1
MAX_GT_SCORE = 0.75
MAX_PHENO_SCORE = 0.25

# Max results matches returned by server.
MAX_RESULTS = 5

# Disclaimer. This text is returned along with match results or server metrics
DISCLAIMER = 'patientMatcher provides data in good faith as a research tool. patientMatcher makes no warranty nor assumes any legal responsibility for any purpose for which the data are used. Users should not attempt in any case to identify patients whose data is returned by the service. Users who intend to publish paper using this software should acknowldge patientMatcher and its developers (https://www.scilifelab.se/facilities/clinical-genomics-stockholm/).'

# Server Host. Used when submitting queries and returning results
MME_HOST = 'https://www.scilifelab.se/facilities/clinical-genomics-stockholm'

# Email notification params.
# Required only if you want to send match notifications to patients contacts
#MAIL_SERVER = mail_port
#MAIL_PORT = email_port
#MAIL_USE_SSL = True or False
#MAIL_USERNAME = 'user_email@mail.se'
#MAIL_PASSWORD = 'mail_password'

# Set NOTIFY_COMPLETE to False if you don't want to notify variants and phenotypes by email
# This way only contact info and matching patients ID will be notified in email body
#NOTIFY_COMPLETE = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

from patientMatcher.parse.patient import gtfeatures_to_genes, gtfeatures_to_variants
LOG = logging.getLogger(__name__)

def match(database, gt_features, max_score):
    """Handles genotype matching algorithm

    Args:
        database(pymongo.database.Database)
        gt_features(list): a list of genomic features (objects)
        max_score(float): a number between 0 and 1

    Returns:
        matches(dict): a dictionary of patient matches with GT score
    """
    matches = {}
    n_gtfeatures = len(gt_features)

    LOG.info('\n\n###### Running genome matcher module ######')

    if n_gtfeatures > 0:

        max_feature_similarity = max_score/n_gtfeatures

        LOG.info('Query patient has {0} genotype features.'.format(n_gtfeatures))
        LOG.info('Each GT feature will contribute with a weight of {0} to a total GT score (max GT score is {1})'.format( max_feature_similarity, max_score ))

        query = {}
        query_fields = []

        genes = gtfeatures_to_genes(gt_features)
        if genes:
            query_fields.append({'genomicFeatures.gene.id' : {"$in" : genes}})

        variants = gtfeatures_to_variants(gt_features)
        if variants:
            query_fields.append({'genomicFeatures.variant': {"$in" : variants} })

        if len(query_fields) > 0:
        # prepare a query that takes into account genes and variants in general (also outside genes!)
            query = { '$or' : query_fields }
            LOG.info('Querying database for genomic features:{}'.format(query))

            # query patients collection
            matching_patients = list(database['patients'].find(query)) # a list of patients with genomic feature/s in one or more of the query genes
            LOG.info("Found {0} matching patients".format(len(matching_patients)))

            # assign a genetic similarity score to each of these patients
            for patient in matching_patients:
                gt_similarity = evaluate_GT_similarity(gt_features, patient['genomicFeatures'], max_feature_similarity)
                match = {
                    'patient_obj' : patient,
                    'geno_score' : gt_similarity,
                }
                matches[patient['_id']] = match

    LOG.info("\n\nFOUND {} patients matching patients's genomic tracts\n\n".format(len(matching_patients)))
    return matches


def evaluate_GT_similarity(query_features, db_patient_features, max_feature_similarity):
    """ Evaluates the genomic similarity of two patients based on genomic similarities

        Args:
            query_patient(list of dictionaries): genomic features of the query patient
            db_patient_features(list of dictionaries): genomic features of a patient in patientMatcher database
            max_similarity(float): a floating point number representing the highest value allowed for a feature

                ## Explanation: for a query patient with one feature max_similarity will be equal to MAX_GT_SCORE
                   For a patient with 2 features max_similarity will be MAX_GT_SCORE/2 and so on.

        Returns:
            patient_similarity(float): the computed genetic similarity among the patients
    """

    matched_features = []
    n_feature = 0

    # loop over the query patient's features
    for feature in query_features:

        matched_features.append(0)
        q_gene = feature['gene'] # query feature's gene id
        q_variant = feature.get('variant', None) # query feature's variant. Not mandatory.

        #loop over the database patient's features:
        for matching_feature in db_patient_features:
            m_gene = matching_feature['gene'] # matching feature's gene id
            m_variant = matching_feature.get('variant') # matching feature's variant. Not mandatory.

            if q_variant and m_variant: # compare only if they have a value. They don't have to be in genes!
                if q_variant == m_variant:
                    matched_features[n_feature] = max_feature_similarity

            elif q_gene and m_gene and matched_features[n_feature] == 0: # Genes not null and no previous variant matching
                if q_gene == m_gene: # matching genes, at least
                    matched_features[n_feature] =  max_feature_similarity/4 #(0.25 of the max_feature_similarity)

        n_feature += 1

    features_sum = sum(matched_features)
    return features_sum



# -*- coding: utf-8 -*-

import logging

LOG = logging.getLogger(__name__)

def patients(database, ids=None):
    """Get all patients in the database

    Args:
        database(pymongo.database.Database)
        ids(list): a list of IDs to return only specified patients

    Returns:
        results(Iterable[dict]): list of patients from mongodb patients collection
    """
    results = None
    query = {}
    if ids: # if only specified patients should be returned
        LOG.info('Querying patients for IDs {}'.format(ids))
        query['_id'] = {'$in' : ids}

    else:
        LOG.info('Return all patients in database')

    results = database['patients'].find(query)
    return results



from patientMatcher.resources import path_to_hpo_terms, path_to_phenotype_annotations

# useful HTTP response status codes with messages
STATUS_CODES = {
    200 : {
        'status_code' : 200
    },
    400 : {
        'message' : 'Invalid request JSON'
    },
    422 : {
        'message' : 'Request does not conform to API specifications'
    },
    401 : {
        'message' : 'Not authorized'
    },
    500 : {
        'message' : 'An error occurred while updating the database'
    },
}

# phenotype terms and annotations are used by phenotype scoring algorithm and
# are updated using the CLI
PHENOTYPE_TERMS= {
    'hpo_ontology' : {
        'url': 'http://purl.obolibrary.org/obo/hp.obo',
        'resource_path' : path_to_hpo_terms
    },
    'hpo_annotations' : {
        'url': 'http://compbio.charite.de/jenkins/job/hpo.annotations/lastStableBuild/artifact/misc/phenotype_annotation_hpoteam.tab',
        'resource_path' : path_to_phenotype_annotations
    }
}



# -*- coding: utf-8 -*-

import logging
from pymongo import MongoClient

LOG = logging.getLogger(__name__)

def delete_by_query(query, mongo_db, mongo_collection):
    """Deletes one or more entries matching a query from a database collection

    Args:
        query(dict): query to be used for deleting the entries
        mongo_db(pymongo.database.Database)
        mongo_collection(str): the name of a collection

    Returns:
        deleted_entries(int): the number of deleted deleted entries or the error
    """

    LOG.info('Removing entries from collection "{0}" using the following parameters:{1}'.format(mongo_collection, query))
    deleted_entries = 0

    try:
        result = mongo_db[mongo_collection].delete_many(query)
        deleted_entries = result.deleted_count
    except Exception as err:
        deleted_entries = err

    return deleted_entries



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree
from patientMatcher import __version__ as version
from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'patientMatcher'
DESCRIPTION = 'a python-based Matchmaker Exchange server'
URL = 'https://github.com/Clinical-Genomics/patientMatcher'
EMAIL = 'chiara.rasi@scilifelab.se'
AUTHOR = 'Chiara Rasi'
KEYWORDS = ['matchmaker exchange', 'genes', 'matchmaker', 'rare diseases', 'genomics', 'genotype', 'phenotype', 'patients', 'genetic disease']
LICENSE = 'MIT'

here = os.path.abspath(os.path.dirname(__file__))

def parse_reqs(req_path='./requirements.txt'):
    """Recursively parse requirements from nested pip files."""
    install_requires = []
    with io.open(os.path.join(here, 'requirements.txt'), encoding='utf-8') as handle:
        # remove comments and empty lines
        lines = (line.strip() for line in handle
                 if line.strip() and not line.startswith('#'))

        for line in lines:
            # check for nested requirements files
            if line.startswith('-r'):
                # recursively call this function
                install_requires += parse_reqs(req_path=line[3:])

            else:
                # add the line as a new requirement
                install_requires.append(line)

    return install_requires

# What packages are required for this module to be executed?
REQUIRED = parse_reqs()

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

# Import the README and use it as the long-description.
# Note: this will only work if 'README.rst' is present in your MANIFEST.in file!
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

# Load the package's __version__.py module as a dictionary.
about = {}
with open(os.path.join(here, NAME, '__version__.py')) as f:
    exec(f.read(), about)


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPi via Twine…')
        os.system('twine upload dist/*')

        sys.exit()


# Where the magic happens:
setup(
    name=NAME,
    version=version,
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    download_url = '/'.join([URL,'tarball',version]),
    keywords = KEYWORDS,
    packages=find_packages(),
    install_requires=REQUIRED,
    include_package_data=True,
    license=LICENSE,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Intended Audience :: Science/Research',
        'Operating System :: MacOS',
        'Operating System :: Unix'
    ],

    entry_points={
        'console_scripts': [
            'pmatcher = patientMatcher.cli.commands:cli'
        ],
    },

    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)



# -*- coding: utf-8 -*-
import os
from pymongo import MongoClient
import logging
from flask import Flask
from flask_mail import Mail
from . import views

logging.basicConfig(level=logging.INFO)
LOG = logging.getLogger(__name__)

def create_app():
    app = None

    try:
        LOG.info('Configuring app from environment variable')
        app = Flask(__name__)
        app.config.from_envvar('PMATCHER_CONFIG')
    except:
        LOG.warning('Environment variable settings not found, configuring from instance file.')
        app_root=os.path.abspath(__file__).split('patientMatcher')[0]

        # check if config file exists under ../instance:
        instance_path = os.path.join(app_root,'patientMatcher', 'instance')
        if not os.path.isfile(os.path.join(instance_path,'config.py')): # running app from tests
            instance_path = os.path.join(app_root,'patientMatcher','patientMatcher','instance')

        app = Flask(__name__, instance_path=instance_path, instance_relative_config=True)
        app.config.from_pyfile('config.py')

    client = MongoClient(app.config['DB_URI'])
    app.client = client
    app.db = client[app.config['DB_NAME']]
    LOG.info('database connection info:{}'.format(app.db))

    if app.config.get('MAIL_SERVER'):
        mail = Mail(app)
        app.mail = mail

    app.register_blueprint(views.blueprint)
    return app



# -*- coding: utf-8 -*-
import pymongo
from patientMatcher.utils.notify import notify_match_external, notify_match_internal, html_format

def test_notify_match_external(match_objs, mock_sender, mock_mail):

    match_obj = match_objs[0] #an external match object with results
    assert match_obj['match_type'] == 'external'

    # When calling the function that sends external match notifications
    notify_complete = True # test notification of complete patient data by email
    notify_match_external(match_obj, mock_sender, mock_mail, notify_complete)

    # make sure send method was called
    assert mock_mail._send_was_called

    # and that mail object message was set correctly
    assert mock_mail._message


def test_notify_match_internal(database, match_objs, mock_sender, mock_mail):

    match_obj = match_objs[2] # an internal match object with results
    assert match_obj['match_type'] == 'internal'

    # insert patient used as query in database:
    assert database['patients'].find().count() == 0
    database['patients'].insert_one({ '_id' : 'external_patient_1'})
    assert database['patients'].find().count() == 1

    # When calling the function that sends internal match notifications
    notify_complete = False # test notification of partial patient data by email
    notify_match_internal(database, match_obj, mock_sender, mock_mail, notify_complete)

    # Test the function that formats the matching results to HTML:
    formatted_results = html_format(match_obj['results'])
    assert '<div style="margin-left: 0em">' in formatted_results

    # make sure send method was called
    assert mock_mail._send_was_called

    # and that mail object message was set correctly
    assert mock_mail._message



# -*- coding: utf-8 -*-

from patientMatcher.utils.add import load_demo
from patientMatcher.parse.patient import mme_patient
from patientMatcher.match.genotype_matcher import match

def test_genotype_matching(demo_data_path, database, json_patients):
    """Testing the genotyping matching algorithm"""

    # load demo data in mock database
    # load demo data to mock database using function located under utils/load
    inserted_ids = load_demo(demo_data_path, database, False)
    assert len(inserted_ids) == 50 # 50 test cases are loaded

    # test conversion to format required for the database:
    test_mme_patients = [ mme_patient(patient) for patient in json_patients]

    # make sure 2 json patient are correctly parsed
    assert len(test_mme_patients) == 2

    # test matching of a patient (with variants in genes) against the demo patients in database
    a_patient = test_mme_patients[0]
    assert a_patient

    # assert patient has genomic features
    gt_features = a_patient['genomicFeatures']
    assert len(gt_features) == 3 # should have 3 features to match

    # match features against database of demo patients:
    matches = match(database, gt_features, 0.5)
    assert len(matches.keys()) == 4 # 4 matching patients are returned

    for key, value in matches.items():
        # patient object should also be returned
        assert 'patient_obj' in value

        # genotype score for each patient should be higher than 0
        assert value['geno_score'] > 0

    # make sure that the algorithm works even if a gene or a variant object is missing:
    # remove gene ID from first gt feature
    gt_features[0]['gene']['id'] = ''
    matches = match(database, gt_features, 0.5)
    # same patients should be returned, because of variant matching instead
    assert len(matches) == 4

    # Remove variant object from second gt feature
    gt_features[1]['variant'] = None
    matches = match(database, gt_features, 0.5)
    # same patients shpuld be returned, because of gene matching instead
    assert len(matches.keys()) == 4
